"""
Provides a suite of native random number generators for use in simulation-based estimation.
Includes uniform, Halton, Latin Hypercube, and normal draws, with optional antithetic and symmetric variants.

Michel Bierlaire
Thu Mar 27 09:46:17 2025
"""

from __future__ import annotations

from typing import Callable, NamedTuple

import numpy as np

from .generators import (
    get_antithetic,
    get_halton_draws,
    get_latin_hypercube_draws,
    get_normal_wichura_draws,
    get_uniform,
)

RandomNumberGenerator = Callable[[int, int], np.ndarray]


class RandomNumberGeneratorTuple(NamedTuple):
    generator: RandomNumberGenerator
    description: str

    @classmethod
    def from_tuple(
        cls, the_tuple: tuple[RandomNumberGenerator, str]
    ) -> RandomNumberGeneratorTuple:
        """
        Create a RandomNumberGeneratorTuple from a legacy tuple.

        :param the_tuple: A tuple of (generator function, description).
        :return: A RandomNumberGeneratorTuple instance.
        """
        return cls(*the_tuple)


def convert_random_generator_tuple(
    the_tuple: RandomNumberGeneratorTuple | tuple[RandomNumberGenerator, str],
) -> RandomNumberGeneratorTuple:
    """
    Convert a random generator specification to a RandomNumberGeneratorTuple.

    :param the_tuple: Either an instance of RandomNumberGeneratorTuple or a legacy tuple.
    :return: A properly typed RandomNumberGeneratorTuple.
    """
    if isinstance(the_tuple, RandomNumberGeneratorTuple):
        return the_tuple
    if not isinstance(the_tuple, tuple):
        raise TypeError(f'Expecting a tuple, not a {type(the_tuple)}')

    return RandomNumberGeneratorTuple.from_tuple(the_tuple)


def uniform_antithetic(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate antithetic uniform random draws.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    return get_antithetic(get_uniform, sample_size, number_of_draws)


def halton2(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate Halton draws with base 2.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    return get_halton_draws(sample_size, number_of_draws, base=2, skip=10)


def halton3(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate Halton draws with base 3.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    return get_halton_draws(sample_size, number_of_draws, base=3, skip=10)


def halton5(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate Halton draws with base 5.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    return get_halton_draws(sample_size, number_of_draws, base=5, skip=10)


def MLHS_anti(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate antithetic Modified Latin Hypercube Sampling draws.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    return get_antithetic(get_latin_hypercube_draws, sample_size, number_of_draws)


def symm_uniform(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate symmetric uniform random draws.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    return get_uniform(sample_size, number_of_draws, symmetric=True)


def symm_uniform_antithetic(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate antithetic symmetric uniform random draws.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    number_local_draws = int(number_of_draws / 2)
    local_draws = symm_uniform(sample_size, number_local_draws)
    return np.concatenate((local_draws, -local_draws), axis=1)


def symm_halton2(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate symmetric Halton draws with base 2.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    return get_halton_draws(
        sample_size, number_of_draws, symmetric=True, base=2, skip=10
    )


def symm_halton3(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate symmetric Halton draws with base 3.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    return get_halton_draws(
        sample_size, number_of_draws, symmetric=True, base=3, skip=10
    )


def symm_halton5(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate symmetric Halton draws with base 5.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    return get_halton_draws(
        sample_size, number_of_draws, symmetric=True, base=5, skip=10
    )


def symm_MLHS(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate symmetric Modified Latin Hypercube Sampling draws.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    return get_latin_hypercube_draws(sample_size, number_of_draws, symmetric=True)


def symm_MLHS_anti(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate antithetic symmetric Modified Latin Hypercube Sampling draws.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    number_local_draws = int(number_of_draws / 2)
    local_draws = symm_MLHS(sample_size, number_local_draws)
    return np.concatenate((local_draws, -local_draws), axis=1)


def normal_antithetic(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate antithetic normal random draws.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    return get_normal_wichura_draws(
        sample_size=sample_size,
        number_of_draws=number_of_draws,
        antithetic=True,
    )


def normal_halton2(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate normal draws from Halton base 2 sequence.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    unif = get_halton_draws(sample_size, number_of_draws, base=2, skip=10)
    return get_normal_wichura_draws(
        sample_size,
        number_of_draws,
        uniform_numbers=unif,
        antithetic=False,
    )


def normal_halton3(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate normal draws from Halton base 3 sequence.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    unif = get_halton_draws(sample_size, number_of_draws, base=3, skip=10)
    return get_normal_wichura_draws(
        sample_size,
        number_of_draws,
        uniform_numbers=unif,
        antithetic=False,
    )


def normal_halton5(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate normal draws from Halton base 5 sequence.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    unif = get_halton_draws(sample_size, number_of_draws, base=5, skip=10)
    return get_normal_wichura_draws(
        sample_size,
        number_of_draws,
        uniform_numbers=unif,
        antithetic=False,
    )


def normal_MLHS(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate normal draws from Modified Latin Hypercube Sampling.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    unif = get_latin_hypercube_draws(sample_size, number_of_draws)
    return get_normal_wichura_draws(
        sample_size,
        number_of_draws,
        uniform_numbers=unif,
        antithetic=False,
    )


def normal_MLHS_anti(sample_size: int, number_of_draws: int) -> np.ndarray:
    """
    Generate antithetic normal draws from Modified Latin Hypercube Sampling.

    :param sample_size: Number of individuals or observations.
    :param number_of_draws: Number of draws per observation.
    :return: A NumPy array of shape (sample_size, number_of_draws).
    """
    unif = get_latin_hypercube_draws(sample_size, int(number_of_draws / 2.0))
    return get_normal_wichura_draws(
        sample_size, number_of_draws, uniform_numbers=unif, antithetic=True
    )


# Dictionary containing native random number generators. Class attribute
native_random_number_generators = {
    'UNIFORM': RandomNumberGeneratorTuple(
        generator=get_uniform, description='Uniform U[0, 1]'
    ),
    'UNIFORM_ANTI': RandomNumberGeneratorTuple(
        generator=uniform_antithetic, description='Antithetic uniform U[0, 1]'
    ),
    'UNIFORM_HALTON2': RandomNumberGeneratorTuple(
        generator=halton2,
        description='Halton draws with base 2, skipping the first 10',
    ),
    'UNIFORM_HALTON3': RandomNumberGeneratorTuple(
        generator=halton3,
        description='Halton draws with base 3, skipping the first 10',
    ),
    'UNIFORM_HALTON5': RandomNumberGeneratorTuple(
        generator=halton5,
        description='Halton draws with base 5, skipping the first 10',
    ),
    'UNIFORM_MLHS': RandomNumberGeneratorTuple(
        generator=get_latin_hypercube_draws,
        description='Modified Latin Hypercube Sampling on [0, 1]',
    ),
    'UNIFORM_MLHS_ANTI': RandomNumberGeneratorTuple(
        generator=MLHS_anti,
        description='Antithetic Modified Latin Hypercube Sampling on [0, 1]',
    ),
    'UNIFORMSYM': RandomNumberGeneratorTuple(
        generator=symm_uniform, description='Uniform U[-1, 1]'
    ),
    'UNIFORMSYM_ANTI': RandomNumberGeneratorTuple(
        generator=symm_uniform_antithetic,
        description='Antithetic uniform U[-1, 1]',
    ),
    'UNIFORMSYM_HALTON2': RandomNumberGeneratorTuple(
        generator=symm_halton2,
        description='Halton draws on [-1, 1] with base 2, skipping the first 10',
    ),
    'UNIFORMSYM_HALTON3': RandomNumberGeneratorTuple(
        generator=symm_halton3,
        description='Halton draws on [-1, 1] with base 3, skipping the first 10',
    ),
    'UNIFORMSYM_HALTON5': RandomNumberGeneratorTuple(
        generator=symm_halton5,
        description='Halton draws on [-1, 1] with base 5, skipping the first 10',
    ),
    'UNIFORMSYM_MLHS': RandomNumberGeneratorTuple(
        generator=symm_MLHS,
        description='Modified Latin Hypercube Sampling on [-1, 1]',
    ),
    'UNIFORMSYM_MLHS_ANTI': RandomNumberGeneratorTuple(
        generator=symm_MLHS_anti,
        description='Antithetic Modified Latin Hypercube Sampling on [-1, 1]',
    ),
    'NORMAL': RandomNumberGeneratorTuple(
        generator=get_normal_wichura_draws, description='Normal N(0, 1) draws'
    ),
    'NORMAL_ANTI': RandomNumberGeneratorTuple(
        generator=normal_antithetic, description='Antithetic normal draws'
    ),
    'NORMAL_HALTON2': RandomNumberGeneratorTuple(
        generator=normal_halton2,
        description='Normal draws from Halton base 2 sequence',
    ),
    'NORMAL_HALTON3': RandomNumberGeneratorTuple(
        generator=normal_halton3,
        description='Normal draws from Halton base 3 sequence',
    ),
    'NORMAL_HALTON5': RandomNumberGeneratorTuple(
        generator=normal_halton5,
        description='Normal draws from Halton base 5 sequence',
    ),
    'NORMAL_MLHS': RandomNumberGeneratorTuple(
        generator=normal_MLHS,
        description='Normal draws from Modified Latin Hypercube Sampling',
    ),
    'NORMAL_MLHS_ANTI': RandomNumberGeneratorTuple(
        generator=normal_MLHS_anti,
        description='Antithetic normal draws from Modified Latin Hypercube Sampling',
    ),
}


def description_of_native_draws() -> dict[str, str]:
    """
    Provides a dictionary of all native draw types and their descriptions.

    :return: Dictionary where keys are generator names and values are human-readable descriptions.
    """
    return {
        key: the_tuple[1] for key, the_tuple in native_random_number_generators.items()
    }



"""
Test the raw estimation results

Michel Bierlaire
Tue Oct 1 19:18:45 2024
"""

import os
import tempfile
import unittest
from datetime import timedelta

from yaml import dump, load, SafeLoader

from biogeme.results_processing import (
    RawEstimationResults,
    serialize_to_yaml,
    deserialize_from_yaml,
)


class TestRawEstimationResults(unittest.TestCase):

    def setUp(self):
        """Set up test fixtures before each test"""
        self.test_data = RawEstimationResults(
            model_name='Test Model',
            user_notes='These are user notes.',
            beta_names=['beta1', 'beta2'],
            beta_values=[0.5, 1.5],
            lower_bounds=[0.0, 1.0],
            upper_bounds=[1.0, 2.0],
            gradient=[0.01, 0.02],
            hessian=[[0.1, 0.2], [0.2, 0.3]],
            bhhh=[[0.01, 0.02], [0.02, 0.03]],
            null_log_likelihood=-150.0,
            initial_log_likelihood=-140.0,
            final_log_likelihood=-120.0,
            data_name='Sample Data',
            sample_size=1000,
            number_of_observations=100,
            monte_carlo=False,
            number_of_draws=500,
            types_of_draws={'type1': 'UNIFORM'},
            number_of_excluded_data=10,
            draws_processing_time=timedelta(seconds=300),
            optimization_messages={'message': 'Optimization successful'},
            convergence=True,
            bootstrap=[[0.5, 1.5], [1.5, 2.5]],
            bootstrap_time=timedelta(seconds=100),
        )

    def test_serialization(self):
        """Test serialization of RawEstimationResults to YAML"""
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            temp_filename = temp_file.name

        # Call the serialize_to_yaml function
        serialize_to_yaml(self.test_data, temp_filename)

        # Verify the content of the file
        with open(temp_filename, 'r') as file:
            content = file.read()
            self.assertIn('Test Model', content)
            self.assertIn('These are user notes.', content)
            self.assertIn('beta1', content)
            self.assertIn('500', content)  # Check number of draws
            self.assertIn('300.0', content)  # Check draws_processing_time as seconds

        # Clean up the temp file
        os.remove(temp_filename)

    def test_deserialization(self):
        """Test deserialization of RawEstimationResults from YAML"""
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            temp_filename = temp_file.name

        # Serialize the data first
        serialize_to_yaml(self.test_data, temp_filename)

        # Deserialize the data
        loaded_data = deserialize_from_yaml(temp_filename)

        # Verify that loaded data matches original data
        self.assertEqual(loaded_data.model_name, self.test_data.model_name)
        self.assertEqual(loaded_data.beta_values, self.test_data.beta_values)
        self.assertEqual(loaded_data.sample_size, self.test_data.sample_size)
        self.assertEqual(loaded_data.number_of_draws, self.test_data.number_of_draws)
        self.assertEqual(
            loaded_data.draws_processing_time, self.test_data.draws_processing_time
        )

        # Clean up the temp file
        os.remove(temp_filename)

    def test_timedelta_representer(self):
        """Test timedelta serialization"""
        result_yaml = dump(
            {
                'draws_processing_time': self.test_data.draws_processing_time.total_seconds()
            }
        )
        self.assertIn('300.0', result_yaml)

    def test_timedelta_constructor(self):
        """Test timedelta deserialization"""
        yaml_data = 'draws_processing_time: "300.0"'
        result = load(yaml_data, Loader=SafeLoader)
        self.assertEqual(result['draws_processing_time'], '300.0')

    def test_empty_serialization(self):
        """Test that serialization works with minimal data"""
        empty_data = RawEstimationResults(
            model_name="",
            user_notes="",
            beta_names=[],
            beta_values=[],
            lower_bounds=[],
            upper_bounds=[],
            gradient=[],
            hessian=[],
            bhhh=[],
            null_log_likelihood=0.0,
            initial_log_likelihood=0.0,
            final_log_likelihood=0.0,
            data_name="",
            sample_size=0,
            number_of_observations=0,
            monte_carlo=False,
            number_of_draws=0,
            types_of_draws={},
            number_of_excluded_data=0,
            draws_processing_time=timedelta(seconds=0),
            optimization_messages={},
            convergence=False,
            bootstrap=[],
            bootstrap_time=None,
        )

        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            temp_filename = temp_file.name

        serialize_to_yaml(empty_data, temp_filename)

        # Verify the file has been written without any errors
        with open(temp_filename, 'r') as file:
            content = file.read()
            self.assertIn('0.0', content)

        # Clean up the temp file
        os.remove(temp_filename)


if __name__ == '__main__':
    unittest.main()



"""Combine several arithmetic expressions and a database to obtain formulas

:author: Michel Bierlaire
:date: Sat Jul 30 12:36:40 2022
"""

from __future__ import annotations

import logging
from typing import Generic, Iterable, NamedTuple, TYPE_CHECKING, TypeVar

import numpy as np

if TYPE_CHECKING:
    from biogeme.database import Database
from biogeme.expressions import (
    Expression,
    TypeOfElementaryExpression,
    Beta,
    Variable,
    RandomVariable,
    Draws,
    list_of_fixed_betas_in_expression,
    list_of_free_betas_in_expression,
    list_of_random_variables_in_expression,
    list_of_draws_in_expression,
)

T = TypeVar('T', bound='Elementary')

try:

    class ElementsTuple(Generic[T], NamedTuple):
        """Data structure for elementary expressions."""

        expressions: dict[str, T] | None
        indices: dict[str, int] | None
        names: list[str]

except TypeError:
    # This exception is raised by Python 3.10.
    class _ElementsTuple(NamedTuple):
        """Data structure for elementary expressions."""

        expressions: dict[str, T] | None
        indices: dict[str, int] | None
        names: list[str]

    class ElementsTuple(Generic[T], _ElementsTuple):
        pass


logger = logging.getLogger(__name__)


def expressions_names_indices(dict_of_elements: dict[str, type[T]]) -> ElementsTuple[T]:
    """Assigns consecutive indices to expressions

    :param dict_of_elements: dictionary of expressions. The keys
        are the names.

    :return: a tuple with the original dictionary, the indices,
        and the sorted names.
    :rtype: ElementsTuple
    """

    indices = {}
    names = sorted(dict_of_elements)

    for i, v in enumerate(names):
        indices[v] = i

    return ElementsTuple(expressions=dict_of_elements, indices=indices, names=names)


class ExpressionRegistry:
    """Class combining managing the ids of an arithmetic expression."""

    def __init__(
        self,
        expressions: Iterable[Expression],
        database: Database,
    ):
        """Ctor

        :param expressions: list of expressions
        :type expressions: list(biogeme.expressions.Expression)

        :param database: database with the variables as column names
        :type database: biogeme.database.Database

        :raises BiogemeError: if an expression contains a variable and
            no database is provided.

        """
        self._expressions: list[Expression] = list(expressions)
        self.database: Database = database
        self._free_betas: list[Beta] | None = None
        self._fixed_betas: list[Beta] | None = None
        self._random_variables: list[RandomVariable] | None = None
        self._draws: list[Draws] | None = None
        self._variables: list[Variable] | None = None
        self._bounds: list[tuple[float, float]] | None = None
        self._require_draws: bool | None = None
        self._free_betas_indices: dict[str, int] | None = None
        self._fixed_betas_indices: dict[str, int] | None = None
        self._variables_indices: dict[str, int] | None = None
        self._random_variables_indices: dict[str, int] | None = None
        self._draws_indices: dict[str, int] | None = None
        self.broadcast()

    @property
    def bounds(self) -> list[tuple[float, float]]:
        if self._bounds is None:
            self._bounds = [
                (beta.lower_bound, beta.upper_bound) for beta in self._free_betas
            ]
        return self._bounds

    @property
    def expressions(self) -> list[Expression]:
        return self._expressions

    @property
    def requires_draws(self) -> bool:
        if self._require_draws is None:
            self._require_draws = any(
                expression.requires_draws() for expression in self.expressions
            )
        return self._require_draws

    @property
    def free_betas_names(self) -> list[str]:
        return [beta.name for beta in self.free_betas]

    @property
    def draws_names(self) -> list[str]:
        return [draw.name for draw in self.draws]

    @property
    def free_betas(self) -> list[Beta]:
        if self._free_betas is None:
            self._free_betas = []
            for f in self.expressions:
                self._free_betas.extend(list_of_free_betas_in_expression(f))
            # Remove duplicates based on the name attribute
            unique_betas = {beta.name: beta for beta in self._free_betas}
            self._free_betas = list(unique_betas.values())
        return self._free_betas

    @property
    def free_betas_indices(self) -> dict[str, int]:
        if self._free_betas_indices is None:
            self._free_betas_indices = {
                beta.name: i for i, beta in enumerate(self.free_betas)
            }
        return self._free_betas_indices

    @property
    def fixed_betas(self) -> list[Beta]:
        if self._fixed_betas is None:
            self._fixed_betas = []
            for f in self.expressions:
                self._fixed_betas.extend(list_of_fixed_betas_in_expression(f))
            # Remove duplicates based on the name attribute
            unique_betas = {beta.name: beta for beta in self._fixed_betas}
            self._fixed_betas = list(unique_betas.values())
        return self._fixed_betas

    @property
    def fixed_betas_indices(self) -> dict[str, int]:
        if self._fixed_betas_indices is None:
            self._fixed_betas_indices = {
                beta.name: i for i, beta in enumerate(self.fixed_betas)
            }
        return self._fixed_betas_indices

    @property
    def random_variables(self) -> list[RandomVariable]:
        if self._random_variables is None:
            self._random_variables = []
            for f in self.expressions:
                self._random_variables.extend(list_of_random_variables_in_expression(f))
            # Remove duplicates based on the name attribute
            unique_rv = {rv.name: rv for rv in self._random_variables}
            self._random_variables = list(unique_rv.values())
        return self._random_variables

    @property
    def random_variables_indices(self) -> dict[str, int]:
        if self._random_variables_indices is None:
            self._random_variables_indices = {
                rv.name: i for i, rv in enumerate(self.random_variables)
            }
        return self._random_variables_indices

    @property
    def draws(self) -> list[Draws]:
        if self._draws is None:
            self._draws = []
            for f in self.expressions:
                self._draws.extend(list_of_draws_in_expression(f))
            # Remove duplicates based on the name attribute
            unique_draws = {draw.name: draw for draw in self._draws}
            self._draws = list(unique_draws.values())
        return self._draws

    @property
    def draws_indices(self) -> dict[str, int]:
        if self._draws_indices is None:
            self._draws_indices = {draw.name: i for i, draw in enumerate(self.draws)}
        return self._draws_indices

    @property
    def variables(self) -> list[Variable]:
        if self._variables is None:
            self._variables = (
                []
                if self.database is None
                else [
                    Variable(name) for name in self.database.dataframe.columns.to_list()
                ]
            )
        return self._variables

    @property
    def variables_indices(self) -> dict[str, int]:
        if self._variables_indices is None:
            self._variables_indices = {
                variable.name: i for i, variable in enumerate(self.variables)
            }
        return self._variables_indices

    @property
    def number_of_free_betas(self) -> int:
        return len(self.free_betas)

    @property
    def number_of_random_variables(self) -> int:
        return len(self.random_variables)

    @property
    def free_betas_init_values(self) -> dict[str, float]:
        """
        Retrieve the initial values of all free beta parameters.

        :return: A dictionary mapping parameter names to their initial values.
        """
        return {
            expression.name: float(expression.init_value)
            for expression in self.free_betas
        }

    def get_betas_array(self, betas_dict: dict[str, float]) -> np.ndarray:
        return np.array([float(betas_dict[beta.name]) for beta in self.free_betas])

    def get_complete_betas_array(self, betas_dict: dict[str, float]) -> np.ndarray:
        complete_dict = self.complete_dict_of_free_beta_values(the_betas=betas_dict)
        return np.array([float(complete_dict[beta.name]) for beta in self.free_betas])

    def get_named_betas_values(self, values: np.ndarray) -> dict[str, float]:
        return {beta.name: float(values[i]) for i, beta in enumerate(self.free_betas)}

    @property
    def list_of_free_betas_init_values(self) -> list[float]:
        """
        Retrieve the initial values of all free beta parameters as a list.

        :return: A list with parameter initial values.
        """
        return list(self.free_betas_init_values.values())

    def complete_dict_of_free_beta_values(
        self, the_betas: dict[str, float]
    ) -> dict[str, float]:
        """Build a list of values for the beta parameters, some of them provided by the user. The others are the
        initial values.

        :param the_betas: user provided values
        :return: A list with parameter  values.
        """
        for beta in self.free_betas:
            if beta.name not in the_betas:
                the_betas[beta.name] = beta.init_value
        return the_betas

    @property
    def fixed_betas_values(self) -> dict[str, float]:
        """
        Retrieve the initial values of all fixed beta parameters.

        :return: A dictionary mapping parameter names to their fixed values.
        """
        return {
            expression.name: expression.init_value for expression in self.fixed_betas
        }

    def draw_types(self) -> dict[str, str]:
        """Retrieve the type of draw for each draw expression"""
        return {expression.name: expression.draw_type for expression in self.draws}

    def broadcast(self) -> None:
        """Broadcast the ids to the expressions"""
        for expression in self.expressions:
            for name, index in self.free_betas_indices.items():
                expression.set_specific_id(
                    name, index, TypeOfElementaryExpression.FREE_BETA
                )
            for name, index in self.fixed_betas_indices.items():
                expression.set_specific_id(
                    name, index, TypeOfElementaryExpression.FIXED_BETA
                )
            for name, index in self.variables_indices.items():
                expression.set_specific_id(
                    name, index, TypeOfElementaryExpression.VARIABLE
                )
            for name, index in self.draws_indices.items():
                expression.set_specific_id(
                    name, index, TypeOfElementaryExpression.DRAWS
                )
            for name, index in self.random_variables_indices.items():
                expression.set_specific_id(
                    name, index, TypeOfElementaryExpression.RANDOM_VARIABLE
                )



"""Implements the probability generating function and the CDF of a
cross-nested logit model. This module is not used by Biogeme
itself. It is essentially used for external use. It has been mainly
implemented to verify the analytical derivatives of these functions.

:author: Michel Bierlaire
:date: Fri Apr 22 09:39:49 2022

"""

from typing import Callable
import numpy as np

from biogeme.expressions import get_dict_values
from biogeme.function_output import FunctionOutput
from biogeme.nests import NestsForCrossNestedLogit
from biogeme.deprecated import deprecated


def cnl_g(
    alternatives: list[int], nests: NestsForCrossNestedLogit
) -> Callable[[np.ndarray], FunctionOutput]:
    """Probability generating function and its derivatives

    :param alternatives: a list of alternatives in a given order. In
        principle, the alternative ids should be integers (to be
        consistent with Biogeme), but it may be actually be any object
        for this specific function.

    :param nests: the object describing the nests.

    :return: function that calculates the G function, and its first
        and second derivatives.
    :rtype: f, g, H = fct(np.array(float))

    """
    order = {alt: index for index, alt in enumerate(alternatives)}
    nbr_of_alternatives = len(alternatives)

    def g_and_deriv(y: np.ndarray) -> FunctionOutput:
        """Probability generating function

        :param y: vector of positive values

        :return: value of the CDF and its derivatives
        :rtype: floap, np.array(float), np.array(np.array(float))
        """

        g = 0.0
        g_i = np.zeros(nbr_of_alternatives)
        g_ij = np.zeros((nbr_of_alternatives, nbr_of_alternatives))
        for m in nests:
            mu_m = m.nest_param
            alphas = get_dict_values(m.dict_of_alpha)
            nest_specific_sum = 0.0
            for alpha_alt, alpha_value in alphas.items():
                if alpha_value != 0 and y[order[alpha_alt]] != 0:
                    nest_specific_sum += (alpha_value * y[order[alpha_alt]]) ** mu_m
            p1 = (1.0 / mu_m) - 1.0
            p2 = (1.0 / mu_m) - 2.0
            g += nest_specific_sum ** (1.0 / mu_m)
            for i in range(nbr_of_alternatives):
                alpha_i = alphas.get(alternatives[i], 0)
                if alpha_i != 0 and y[i] != 0:
                    g_i[i] += alpha_i**mu_m * y[i] ** (mu_m - 1) * nest_specific_sum**p1
                    g_ij[i][i] += (1 - mu_m) * nest_specific_sum**p2 * alpha_i ** (
                        2 * mu_m
                    ) * y[i] ** (2 * mu_m - 2.0) + (
                        mu_m - 1
                    ) * nest_specific_sum**p1 * alpha_i**mu_m * y[
                        i
                    ] ** (
                        mu_m - 2
                    )
                    for j in range(i + 1, nbr_of_alternatives):
                        alpha_j = alphas.get(alternatives[j], 0)
                        if alpha_j != 0 and y[j] != 0:
                            g_ij[i][j] += (
                                (1 - mu_m)
                                * nest_specific_sum**p2
                                * (alpha_i * alpha_j) ** mu_m
                                * (y[i] * y[j]) ** (mu_m - 1.0)
                            )

        for i in range(nbr_of_alternatives):
            for j in range(i + 1, nbr_of_alternatives):
                g_ij[j][i] = g_ij[i][j]

        return FunctionOutput(function=g, gradient=g_i, hessian=g_ij)

    return g_and_deriv


@deprecated(new_func=cnl_g)
def cnl_G(
    alternatives: list[int], nests: NestsForCrossNestedLogit
) -> Callable[[np.ndarray], FunctionOutput]:
    pass


def cnl_cdf(
    alternatives: list[int], nests: NestsForCrossNestedLogit
) -> Callable[[np.ndarray], FunctionOutput]:
    """Cumulative distribution function and its derivatives

    :param alternatives: a list of alternatives in a given order. In
        principle, the alternative ids should be integers (to be
        consistent with Biogeme), but it may be actually be any object
        for this specific function.

    :param nests: a tuple containing as many items as nests.

    :return: function that calculates the CDF, and its first
        and second derivatives.

    """
    nbr_of_alternatives = len(alternatives)

    g_fct = cnl_g(alternatives, nests)

    def f_and_deriv(xi: np.ndarray) -> FunctionOutput:
        """Cumulative distribution function

        :param xi: vector of arguments
        :type xi: np.array(float)

        :return: value of the CDF and its derivatives
        :rtype: float, np.array(float), np.array(np.array(float))
        """
        y = np.where(xi == np.inf, 0, np.exp(-xi))
        g_output: FunctionOutput = g_fct(y)
        g = g_output.function
        g_i = g_output.gradient
        g_ii = g_output.hessian

        f = np.exp(-g)
        f_i = g_i * y * f
        f_ij = np.zeros((nbr_of_alternatives, nbr_of_alternatives))

        for i in range(nbr_of_alternatives):
            f_ij[i][i] = (
                f * y[i] * y[i] * (g_i[i] * g_i[i] - g_ii[i][i]) - f * g_i[i] * y[i]
            )
            for j in range(i + 1, nbr_of_alternatives):
                f_ij[i][j] = f * y[i] * y[j] * (g_i[i] * g_i[j] - g_ii[i][j])
        for i in range(nbr_of_alternatives):
            for j in range(i + 1, nbr_of_alternatives):
                f_ij[j][i] = f_ij[i][j]

        return FunctionOutput(function=f, gradient=f_i, hessian=f_ij)

    return f_and_deriv


@deprecated(new_func=cnl_cdf)
def cnl_CDF(
    alternatives: list[int], nests: NestsForCrossNestedLogit
) -> Callable[[np.ndarray], FunctionOutput]:
    pass



"""Defines  a segmentation catalog

Michel Bierlaire
Wed Apr 16 18:35:02 2025

"""

from __future__ import annotations

import logging
from itertools import product

from biogeme.exceptions import BiogemeError
from biogeme.expressions import Beta, Expression, NamedExpression
from biogeme.segmentation import DiscreteSegmentationTuple, Segmentation

from .catalog import Catalog
from .configuration import SELECTION_SEPARATOR, SEPARATOR
from .controller import Controller

logger = logging.getLogger(__name__)


def segmentation_catalogs(
    generic_name: str,
    beta_parameters: list[Beta],
    potential_segmentations: tuple[DiscreteSegmentationTuple, ...],
    maximum_number: int,
) -> list[Catalog]:
    """Generate catalogs for potential segmentations of a parameter

    :param generic_name: name used for the definition of the group of catalogs
    :param beta_parameters: list of parameters to be segmented
    :param potential_segmentations: tuple of potential segmentations
    :param maximum_number: maximum number of segmentations to consider
    """

    for segmentation in potential_segmentations:
        for key, value in segmentation.mapping.items():
            if SEPARATOR in value or SELECTION_SEPARATOR in value:
                error_msg = (
                    f'Invalid segment name for variable {segmentation.variable.name}='
                    f'{key}: [{value}]. Characters [{SEPARATOR}] and '
                    f'[{SELECTION_SEPARATOR}] are reserved for specification coding.'
                )
                raise BiogemeError(error_msg)

    def get_name_from_combination(combination: tuple[bool, ...]) -> str:
        """Assign a name to a combination"""
        if sum(combination) == 0:
            return 'no_seg'

        return '-'.join(
            [
                segment.variable.name
                for keep, segment in zip(combination, potential_segmentations)
                if keep
            ]
        )

    def get_expression_from_combination(
        the_beta_parameter: Beta, combination: tuple[bool, ...]
    ) -> Expression:
        """Assign an expression to a combination"""
        if sum(combination) == 0:
            # No segmentation
            return the_beta_parameter

        selected_expressions = (
            segment
            for keep, segment in zip(combination, potential_segmentations)
            if keep
        )

        the_segmentation = Segmentation(the_beta_parameter, selected_expressions)
        return the_segmentation.segmented_beta()

    if not isinstance(beta_parameters, list):
        error_msg = (
            f'A list is expected for beta_parameters, and not an object of type '
            f'{type(beta_parameters)}'
        )
        raise BiogemeError(error_msg)

    list_of_possibilities = [
        combination
        for combination in product([False, True], repeat=len(potential_segmentations))
        if sum(combination) <= maximum_number
    ]
    catalogs = []
    names = [
        get_name_from_combination(combination) for combination in list_of_possibilities
    ]
    the_controller = Controller(controller_name=generic_name, specification_names=names)
    for beta_parameter in beta_parameters:
        named_expressions = [
            NamedExpression(
                name=get_name_from_combination(combination),
                expression=get_expression_from_combination(beta_parameter, combination),
            )
            for combination in list_of_possibilities
        ]
        name = f'segmented_{beta_parameter.name}'
        catalog = Catalog(
            catalog_name=name,
            named_expressions=named_expressions,
            controlled_by=the_controller,
        )
        catalogs.append(catalog)
    return catalogs


class SegmentedParameters:
    """Class managing the names of segmented and alternative specific parameters"""

    def __init__(self, beta_parameters: list[Beta], alternatives: tuple[str, ...]):
        """Constructor"""

        # The parameters are organized as follows:
        #   - all generic parameters,
        #   - all parameters associated with the first alternative,
        #   - all parameters associated with the second alternative,
        #   - etc.
        self.beta_parameters: list[Beta] = beta_parameters
        self.all_parameters: list[Beta] = beta_parameters.copy()
        self.alternatives: tuple[str, ...] = alternatives
        for alternative in self.alternatives:
            self.all_parameters += [
                Beta(
                    name=f'{beta.name}_{alternative}',
                    value=beta.init_value,
                    lowerbound=beta.lower_bound,
                    upperbound=beta.upper_bound,
                    status=beta.status,
                )
                for beta in beta_parameters
            ]

    def get_index(self, beta_index: int, alternative: str | None):
        """Returns the index in the list of the Beta parameter with
            the given index specific to the given alternative

        :param beta_index: index of the Beta in the generic list
        :type beta_index: int

        :param alternative: name of the alternative, or None for the generic parameter
        :type alternative: str or None

        """
        if alternative is None:
            return beta_index
        alt_index = self.alternatives.index(alternative)
        return beta_index + (alt_index + 1) * len(self.beta_parameters)

    def get_beta(self, beta_index: int, alternative: str | None):
        """Return the Beta parameters for the given index and given alternative

        :param beta_index: index of the Beta in the generic list
        :type beta_index: int

        :param alternative: name of the alternative, or None for the generic parameter
        :type alternative: str or None

        """
        return self.all_parameters[self.get_index(beta_index, alternative)]



"""
Representation of an ellipse

Michel Bierlaire
Sat Dec 21 16:45:00 2024
"""

from dataclasses import dataclass

from matplotlib import pyplot as plt
from matplotlib.patches import Ellipse as MatplotEllipse
import numpy as np


@dataclass
class Ellipse:
    """Contains the information needed to draw an ellipse"""

    center_x: float
    center_y: float
    sin_phi: float
    cos_phi: float
    axis_one: float
    axis_two: float


def draw_ellipse(ellipse: Ellipse, ax=None, **kwargs):
    """
    Draws an ellipse using matplotlib.

    :param ellipse: An Ellipse object containing the parameters of the ellipse.
    :param ax: A matplotlib Axes object. If None, a new figure and axes will be created.
    :param kwargs: Additional keyword arguments passed to matplotlib.patches.Ellipse.
    """
    # Compute the rotation angle of the ellipse in degrees
    phi = np.arctan2(ellipse.sin_phi, ellipse.cos_phi)
    angle = np.degrees(phi)

    # Create the matplotlib ellipse
    patch = MatplotEllipse(
        (ellipse.center_x, ellipse.center_y),  # Center of the ellipse
        width=2 * ellipse.axis_one,  # Major axis (diameter)
        height=2 * ellipse.axis_two,  # Minor axis (diameter)
        angle=angle,  # Rotation angle in degrees
        **kwargs  # Additional styling options
    )

    # Add the ellipse to the axes
    if ax is None:
        fig, ax = plt.subplots(figsize=(6, 6))
    ax.add_patch(patch)

    # Set limits for better visualization
    ax.set_xlim(
        ellipse.center_x - 1.5 * ellipse.axis_one,
        ellipse.center_x + 1.5 * ellipse.axis_one,
    )
    ax.set_ylim(
        ellipse.center_y - 1.5 * ellipse.axis_two,
        ellipse.center_y + 1.5 * ellipse.axis_two,
    )
    ax.set_aspect('equal', adjustable='datalim')

    # Add labels and grid
    ax.axhline(0, color='gray', linestyle='--', linewidth=0.5)
    ax.axvline(0, color='gray', linestyle='--', linewidth=0.5)
    ax.set_xlabel('X-axis')
    ax.set_ylabel('Y-axis')
    ax.grid(True)

    # Show plot if ax was None
    if ax is None:
        plt.show()



"""Class representing the characterization of a discrete segmentation.

A discrete segmentation divides a population based on the values of a socio-economic
variable, mapping each value to a category name. A reference category is defined
for comparison in model estimation.

Michel Bierlaire
Thu Apr 3 10:08:10 2025
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from biogeme.exceptions import BiogemeError

if TYPE_CHECKING:
    from biogeme.expressions import Variable


class DiscreteSegmentationTuple:
    """Characterization of a discrete segmentation.

    This class is used to define how a discrete socio-economic variable
    is mapped to category names for segmentation purposes. One category
    is selected as the reference for use in estimation.
    """

    def __init__(
        self,
        variable: Variable | str,
        mapping: dict[int, str],
        reference: str | None = None,
    ):
        """Initialize a discrete segmentation characterization.

        :param variable: Socio-economic variable used for segmentation, either as a Variable instance or a string name.
        :param mapping: Dictionary mapping integer values of the variable to category names.
        :param reference: Optional name of the reference category. If not provided, the first category in the mapping is used.

        :raises BiogemeError: If the provided reference category is not found in the mapping.
        """
        from biogeme.expressions import Variable

        self.variable: Variable = (
            variable if isinstance(variable, Variable) else Variable(variable)
        )
        self.mapping: dict[int, str] = mapping
        if reference is None:
            self.reference: str = next(iter(mapping.values()))
        elif reference not in mapping.values():
            error_msg = (
                f'Reference category {reference} does not appear in the list '
                f'of categories: {mapping.values()}'
            )
            raise BiogemeError(error_msg)
        else:
            self.reference: str = reference

        if self.reference is None:
            raise BiogemeError('Reference should not be None')

    def __repr__(self) -> str:
        result = f'{self.variable.name}: [{self.mapping}] ref: {self.reference}'
        return result

    def __str__(self) -> str:
        result = f'{self.variable.name}: [{self.mapping}] ref: {self.reference}'
        return result



"""
Module in charge of the actual calculation of the formula on the database.

Michel Bierlaire
Wed Mar 26 19:30:57 2025
"""

from __future__ import annotations

import logging

import jax
import jax.numpy as jnp
import numpy as np

from biogeme.database import Database
from biogeme.exceptions import BiogemeError
from biogeme.expressions import (
    Expression,
    build_vectorized_function,
    collect_init_values,
)
from biogeme.floating_point import JAX_FLOAT, NUMPY_FLOAT
from biogeme.function_output import FunctionOutput, NamedFunctionOutput
from biogeme.model_elements import ModelElements
from biogeme.second_derivatives import SecondDerivativesMode

logger = logging.getLogger(__name__)


class CompiledFormulaEvaluator:
    """
    Compiles and evaluates a Biogeme expression using JAX for efficient
        repeated computation.
    """

    def __init__(
        self,
        model_elements: ModelElements,
        second_derivatives_mode: SecondDerivativesMode,
        numerically_safe: bool,
    ):
        """
        Prepares and compiles the JAX function for evaluating a Biogeme expression.

        :param model_elements: All elements needed to calculate the expression.
        :param second_derivatives_mode: specifies how second derivatives are calculated.
        :param numerically_safe: improves the numerical stability of the calculations.
        """
        from biogeme.expressions import build_vectorized_function

        self.model_elements = model_elements
        self.second_derivatives_mode = second_derivatives_mode
        self.numerically_safe = numerically_safe
        self.free_betas_names = (
            self.model_elements.expressions_registry.free_betas_names
        )
        self.data_jax = (
            self.model_elements.database.data_jax
            if self.model_elements.database is not None
            else None
        )
        self.draws_jax = (
            self.model_elements.draws_management.draws_jax
            if self.model_elements.draws_management is not None
            else None
        )
        n_rv = self.model_elements.expressions_registry.number_of_random_variables
        self.random_variables_jax = jnp.zeros((n_rv,), dtype=JAX_FLOAT)

        log_likelihood = self.model_elements.loglikelihood
        if log_likelihood is None:
            error_message = (
                f'No expression found for log likelihood. '
                f'Available expressions: {self.model_elements.formula_names}'
            )
            raise BiogemeError(error_message)
        the_function = log_likelihood.recursive_construct_jax_function(
            numerically_safe=self.numerically_safe
        )
        vectorized_function = build_vectorized_function(the_function)

        if self.model_elements.weight is not None:
            weight_function = (
                self.model_elements.weight.recursive_construct_jax_function(
                    numerically_safe=numerically_safe
                )
            )
            vectorized_weight_function = build_vectorized_function(weight_function)
        else:
            vectorized_weight_function = None

        def sum_function(
            params: list[float],
            data: jnp.ndarray,
            draws: jnp.ndarray,
            random_variables: jnp.ndarray,
        ) -> tuple[jnp.ndarray, jnp.ndarray]:
            values = vectorized_function(params, data, draws, random_variables)
            if vectorized_weight_function is not None:
                weights = vectorized_weight_function(
                    params, data, draws, random_variables
                )
                values *= weights
            return jnp.asarray(jnp.sum(values), dtype=JAX_FLOAT), values

        self.sum_function = jax.jit(sum_function)

    def evaluate(
        self,
        the_betas: dict[str, float],
        gradient: bool,
        hessian: bool,
        bhhh: bool,
    ) -> FunctionOutput:
        free_betas_values = (
            self.model_elements.expressions_registry.get_complete_betas_array(
                betas_dict=the_betas
            )
        )

        if not gradient:
            return self._evaluate_function_only(free_betas_values)

        if bhhh:
            if hessian:
                return self._evaluate_autodiff_hessian_bhhh(free_betas_values)
            else:
                return self._evaluate_bhhh_only(free_betas_values)

        if hessian:
            return self._evaluate_autodiff_hessian(free_betas_values)

        return self._evaluate_function_and_gradient(free_betas_values)

    def _evaluate_function_only(self, free_betas_values):
        value_jax, _ = self.sum_function(
            free_betas_values,
            self.data_jax,
            self.draws_jax,
            self.random_variables_jax,
        )
        return FunctionOutput(
            function=float(value_jax),
            gradient=None,
            hessian=None,
            bhhh=None,
        )

    def _evaluate_function_and_gradient(self, free_betas_values):
        value_and_grad_fn = jax.jit(
            jax.value_and_grad(
                lambda p, d, r, rv: self.sum_function(p, d, r, rv)[0], argnums=0
            )
        )
        value, the_gradient = value_and_grad_fn(
            free_betas_values,
            self.data_jax,
            self.draws_jax,
            self.random_variables_jax,
        )
        return FunctionOutput(
            function=float(value),
            gradient=np.asarray(the_gradient, dtype=NUMPY_FLOAT),
            hessian=None,
            bhhh=None,
        )

    def _evaluate_autodiff_hessian(self, free_betas_values):
        if self.second_derivatives_mode == SecondDerivativesMode.NEVER:
            error_msg = 'The second derivatives are not supposed to be evaluated'
            raise BiogemeError(error_msg)

        value_and_grad_fn = jax.jit(
            jax.value_and_grad(
                lambda p, d, r, rv: self.sum_function(p, d, r, rv)[0],
                argnums=0,
            )
        )
        value, the_gradient = value_and_grad_fn(
            free_betas_values,
            self.data_jax,
            self.draws_jax,
            self.random_variables_jax,
        )
        if jnp.all(the_gradient == 0.0):
            the_hessian = np.zeros(
                (len(free_betas_values), len(free_betas_values)), dtype=NUMPY_FLOAT
            )
        elif self.second_derivatives_mode == SecondDerivativesMode.FINITE_DIFFERENCES:
            the_hessian = self._evaluate_finite_difference_hessian(free_betas_values)
        else:
            hessian_fn = jax.jit(
                jax.jacfwd(
                    jax.grad(
                        lambda p, d, r, rv: self.sum_function(p, d, r, rv)[0],
                        argnums=0,
                    ),
                    argnums=0,
                )
            )
            hess_autodiff = hessian_fn(
                free_betas_values,
                self.data_jax,
                self.draws_jax,
                self.random_variables_jax,
            )
            if jnp.any(jnp.isnan(hess_autodiff)):
                raise BiogemeError(
                    'The calculation of second derivatives generated numerical errors.'
                )
            the_hessian = np.asarray(hess_autodiff, dtype=NUMPY_FLOAT)

        return FunctionOutput(
            function=float(value),
            gradient=np.asarray(the_gradient, dtype=NUMPY_FLOAT),
            hessian=the_hessian,
            bhhh=None,
        )

    def _evaluate_bhhh_only(self, free_betas_values):
        _, individual_values = self.sum_function(
            free_betas_values,
            self.data_jax,
            self.draws_jax,
            self.random_variables_jax,
        )

        def one_gradient(p, d, r, rv):
            loglik_fn = (
                self.model_elements.loglikelihood.recursive_construct_jax_function(
                    numerically_safe=self.numerically_safe
                )
            )
            vectorized_function = build_vectorized_function(loglik_fn)
            draw_values = vectorized_function(p, d[None, :], r[None, :, :], rv)
            return jnp.mean(draw_values)

        per_obs_grad_fn = jax.jit(
            jax.vmap(jax.grad(one_gradient, argnums=0), in_axes=(None, 0, 0, 0))
        )
        free_betas_values_jnp = jnp.asarray(free_betas_values, dtype=JAX_FLOAT)
        random_variables_broadcast = jnp.tile(
            self.random_variables_jax[None, :], (self.data_jax.shape[0], 1)
        )
        individual_gradients = per_obs_grad_fn(
            free_betas_values_jnp,
            self.data_jax,
            self.draws_jax,
            random_variables_broadcast,
        )

        expected_shape = (self.data_jax.shape[0], len(free_betas_values))
        if individual_gradients.shape != expected_shape:
            individual_gradients = jnp.tile(
                individual_gradients, (self.data_jax.shape[0], 1)
            )

        the_gradient = jnp.sum(individual_gradients, axis=0)
        bhhh_matrix = jnp.sum(
            jnp.stack([jnp.outer(g, g) for g in individual_gradients]), axis=0
        )

        return FunctionOutput(
            function=float(jnp.sum(individual_values)),
            gradient=np.asarray(the_gradient, dtype=NUMPY_FLOAT),
            hessian=None,
            bhhh=np.asarray(bhhh_matrix, dtype=NUMPY_FLOAT),
        )

    def _evaluate_autodiff_hessian_bhhh(self, free_betas_values):
        bhhh_result = self._evaluate_bhhh_only(free_betas_values)
        hessian = self._evaluate_autodiff_hessian(free_betas_values).hessian
        return FunctionOutput(
            function=bhhh_result.function,
            gradient=bhhh_result.gradient,
            hessian=hessian,
            bhhh=bhhh_result.bhhh,
        )

    def _evaluate_finite_difference_hessian(self, free_betas_values):
        import scipy.optimize as so

        def func_for_fd(betas_array):
            return float(
                self.sum_function(
                    betas_array,
                    self.data_jax,
                    self.draws_jax,
                    self.random_variables_jax,
                )[0]
            )

        eps = np.sqrt(np.finfo(float).eps)
        n = len(free_betas_values)
        the_hessian = np.zeros((n, n), dtype=NUMPY_FLOAT)
        for i in range(n):
            x0 = np.array(free_betas_values)
            ei = np.zeros_like(x0)
            ei[i] = eps
            g_plus = so.approx_fprime(x0 + ei, func_for_fd, eps)
            g_minus = so.approx_fprime(x0 - ei, func_for_fd, eps)
            the_hessian[i, :] = (g_plus - g_minus) / (2 * eps)
        return the_hessian

    def evaluate_individual(
        self,
        the_betas: dict[str, float],
    ) -> np.ndarray:
        """
        Evaluates the compiled expression using provided beta values and returns
        the value of the expression for each observation.

        :param the_betas: Dictionary of parameter names to values.
        :return: A numpy array with one value per observation.
        """
        free_betas_values = (
            self.model_elements.expressions_registry.get_complete_betas_array(
                betas_dict=the_betas
            )
        )

        _, individual_values = self.sum_function(
            free_betas_values, self.data_jax, self.draws_jax, self.random_variables_jax
        )
        return np.asarray(individual_values, dtype=NUMPY_FLOAT)


def calculate_single_formula(
    model_elements: ModelElements,
    the_betas: dict[str, float],
    gradient: bool,
    hessian: bool,
    bhhh: bool,
    second_derivatives_mode: SecondDerivativesMode,
    numerically_safe: bool,
) -> FunctionOutput:
    """
    Evaluates a single Biogeme expression using JAX, optionally computing the gradient
        and Hessian.

    :param model_elements: All elements needed to calculate the expression.
    :param the_betas: Dictionary of parameter names to values.
    :param gradient: If True, compute the gradient.
    :param hessian: If True, compute the Hessian (requires gradient=True).
    :param bhhh: Unused here, included for compatibility.
    :param second_derivatives_mode: specifies how second derivatives are calculated.
    :param numerically_safe: improves the numerical stability of the calculations.
    :return: A BiogemeFunctionOutput with the value, gradient,
       and optionally the Hessian.
    """
    the_compiled_formula = CompiledFormulaEvaluator(
        model_elements=model_elements,
        second_derivatives_mode=second_derivatives_mode,
        numerically_safe=numerically_safe,
    )
    return the_compiled_formula.evaluate(
        the_betas=the_betas, gradient=gradient, hessian=hessian, bhhh=bhhh
    )


def calculate_single_formula_from_expression(
    expression: Expression,
    database: Database,
    number_of_draws: int,
    the_betas: dict[str, float],
    second_derivatives_mode: SecondDerivativesMode,
    numerically_safe: bool,
) -> float:
    model_elements = ModelElements.from_expression_and_weight(
        log_like=expression,
        weight=None,
        database=database,
        number_of_draws=number_of_draws,
    )
    result = calculate_single_formula(
        model_elements=model_elements,
        second_derivatives_mode=second_derivatives_mode,
        numerically_safe=numerically_safe,
        the_betas=the_betas,
        gradient=False,
        hessian=False,
        bhhh=False,
    )
    return result.function


def evaluate_formula(
    model_elements: ModelElements,
    the_betas: dict[str, float],
    second_derivatives_mode: SecondDerivativesMode,
    numerically_safe: bool,
) -> float:
    """
    Evaluates a single Biogeme expression using JAX.

    :param model_elements: All elements needed to calculate the expression.
    :param the_betas: Dictionary of parameter names to values.
    :param second_derivatives_mode: specifies how second derivatives are calculated.
    :param numerically_safe: improves the numerical stability of the calculations.
    :return: the value of the expression.
    """
    result = calculate_single_formula(
        model_elements=model_elements,
        the_betas=the_betas,
        gradient=False,
        hessian=False,
        bhhh=False,
        second_derivatives_mode=second_derivatives_mode,
        numerically_safe=numerically_safe,
    )
    return result.function


def evaluate_model_per_row(
    model_elements: ModelElements,
    the_betas: dict[str, float],
    second_derivatives_mode: SecondDerivativesMode,
    numerically_safe: bool,
) -> np.ndarray:
    """
    Evaluates a Biogeme expression for each entry in the database and returns
    individual results.

    This function compiles the expression using JAX, applies it to all observations
    in the database, and returns a NumPy array containing the evaluated values
    per observation. The result is not aggregated or summed.

    :param model_elements: All elements needed to calculate the expression.
    :param the_betas: Dictionary mapping parameter names to their values.
    :param second_derivatives_mode: specifies how second derivatives are calculated.
    :param numerically_safe: improves the numerical stability of the calculations.
    :return: A NumPy array of values, one for each observation in the database.
    """
    the_compiled_formula = CompiledFormulaEvaluator(
        model_elements=model_elements,
        second_derivatives_mode=second_derivatives_mode,
        numerically_safe=numerically_safe,
    )
    return the_compiled_formula.evaluate_individual(the_betas=the_betas)


def evaluate_expression(
    expression: Expression,
    numerically_safe: bool,
    database: Database | None = None,
    betas: dict[str, float] | None = None,
    number_of_draws: int = 1000,
    aggregation: bool = False,
) -> np.ndarray | float:
    """Evaluate an arithmetic expression

    :param expression: the expression to be evaluated
    :param numerically_safe: if True, the numerical stability of the evaluation is improved, possibly at the expense
         of calculation speed. Set it to False except if necessary.
    :param database: database, needed if the expression involves `Variable`
    :param betas: values of the parameters, if the expression involves `Beta`
    :param number_of_draws: number of draws for Monte Carlo integration, if the expression involves it.
    :param aggregation: if True, the sum over all rows is calculated. If False, the value for each row is returned.
    """
    if database is None:
        database = Database.dummy_database()
    model_elements = ModelElements.from_expression_and_weight(
        log_like=expression,
        weight=None,
        database=database,
        number_of_draws=number_of_draws,
    )
    if betas is None:
        betas = collect_init_values(expression=expression)
    if aggregation:
        return evaluate_formula(
            model_elements=model_elements,
            the_betas=betas,
            second_derivatives_mode=SecondDerivativesMode.NEVER,
            numerically_safe=numerically_safe,
        )

    return evaluate_model_per_row(
        model_elements=model_elements,
        the_betas=betas,
        second_derivatives_mode=SecondDerivativesMode.NEVER,
        numerically_safe=numerically_safe,
    )


def get_value_and_derivatives(
    expression: Expression,
    numerically_safe: bool,
    betas: dict[str, float] | None = None,
    database: Database | None = None,
    number_of_draws: int = 1000,
    gradient: bool = True,
    hessian: bool = True,
    bhhh: bool = True,
    named_results: bool = False,
) -> FunctionOutput | NamedFunctionOutput:
    if database is None:
        from biogeme.database import Database

        database = Database.dummy_database()
    model_elements = ModelElements.from_expression_and_weight(
        log_like=expression,
        weight=None,
        database=database,
        number_of_draws=number_of_draws,
    )

    the_compiled_formula = CompiledFormulaEvaluator(
        model_elements=model_elements,
        second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
        numerically_safe=numerically_safe,
    )
    if betas is None:
        betas = collect_init_values(expression=expression)
    result: FunctionOutput = the_compiled_formula.evaluate(
        the_betas=betas, gradient=gradient, hessian=hessian, bhhh=bhhh
    )
    if not named_results:
        return result
    named_results = NamedFunctionOutput(
        function_output=result,
        mapping=model_elements.expressions_registry.free_betas_indices,
    )
    return named_results


def get_value_c(
    expression: Expression,
    numerically_safe: bool,
    database: Database | None = None,
    betas: dict[str, float] | None = None,
    number_of_draws: int = 1000,
    aggregation: bool = False,
) -> np.ndarray | float:
    """For backward compatibility. This function used to be a member of the
    Expression class."""
    return evaluate_expression(
        expression=expression,
        numerically_safe=numerically_safe,
        database=database,
        betas=betas,
        number_of_draws=number_of_draws,
        aggregation=aggregation,
    )



"""Arithmetic expressions accepted by Biogeme: comparison operators

Michel Bierlaire
Wed Mar 26 13:17:40 2025
"""

from __future__ import annotations

import logging

from jax import Array, lax

from .base_expressions import ExpressionOrNumeric
from .binary_expressions import BinaryOperator
from .jax_utils import JaxFunctionType

logger = logging.getLogger(__name__)


class ComparisonOperator(BinaryOperator):
    """Base class for comparison expressions."""

    def __init__(self, left: ExpressionOrNumeric, right: ExpressionOrNumeric):
        """Constructor

        :param left: first arithmetic expression
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression
        :type right: biogeme.expressions.Expression
        """
        super().__init__(left, right)


class Equal(ComparisonOperator):
    """
    Logical equal
    """

    def __init__(self, left: ExpressionOrNumeric, right: ExpressionOrNumeric):
        """Constructor

        :param left: first arithmetic expression
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression
        :type right: biogeme.expressions.Expression
        """
        super().__init__(left, right)

    def deep_flat_copy(self) -> Equal:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        left_copy = self.left.deep_flat_copy()
        right_copy = self.right.deep_flat_copy()
        return type(self)(left=left_copy, right=right_copy)

    def __str__(self) -> str:
        return f'({self.left} == {self.right})'

    def __repr__(self) -> str:
        return f'Equal({repr(self.left)}, {repr(self.right)})'

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        r = 1 if self.left.get_value() == self.right.get_value() else 0
        return r

    def recursive_construct_jax_function(self, numerically_safe: bool):
        import jax.numpy as jnp

        left_fn = self.left.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )
        right_fn = self.right.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ):
            is_equal = left_fn(
                parameters, one_row, the_draws, the_random_variables
            ) == right_fn(parameters, one_row, the_draws, the_random_variables)
            return lax.cond(is_equal, lambda _: 1.0, lambda _: 0.0, operand=None)

        return the_jax_function


class NotEqual(ComparisonOperator):
    """
    Logical not equal
    """

    def __init__(self, left: ExpressionOrNumeric, right: ExpressionOrNumeric):
        """Constructor

        :param left: first arithmetic expression
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression
        :type right: biogeme.expressions.Expression
        """
        super().__init__(left, right)

    def deep_flat_copy(self) -> NotEqual:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        left_copy = self.left.deep_flat_copy()
        right_copy = self.right.deep_flat_copy()
        return type(self)(left=left_copy, right=right_copy)

    def __str__(self) -> str:
        return f'({self.left} != {self.right})'

    def __repr__(self) -> str:
        return f'NotEqual({repr(self.left)}, {repr(self.right)})'

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        r = 1 if self.left.get_value() != self.right.get_value() else 0
        return r

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        import jax.numpy as jnp

        left_fn: JaxFunctionType = self.left.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )
        right_fn: JaxFunctionType = self.right.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> Array:
            is_not_equal = left_fn(
                parameters, one_row, the_draws, the_random_variables
            ) != right_fn(parameters, one_row, the_draws, the_random_variables)
            return lax.cond(is_not_equal, lambda _: 1.0, lambda _: 0.0, operand=None)

        return the_jax_function


class LessOrEqual(ComparisonOperator):
    """
    Logical less or equal
    """

    def __init__(self, left: ExpressionOrNumeric, right: ExpressionOrNumeric):
        """Constructor

        :param left: first arithmetic expression
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression
        :type right: biogeme.expressions.Expression

        """
        super().__init__(left, right)

    def deep_flat_copy(self) -> LessOrEqual:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        left_copy = self.left.deep_flat_copy()
        right_copy = self.right.deep_flat_copy()
        return type(self)(left=left_copy, right=right_copy)

    def __str__(self) -> str:
        return f'({self.left} <= {self.right})'

    def __repr__(self) -> str:
        return f'LessOrEqual({repr(self.left)}, {repr(self.right)})'

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        r = 1 if self.left.get_value() <= self.right.get_value() else 0
        return r

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        import jax.numpy as jnp

        left_fn: JaxFunctionType = self.left.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )
        right_fn: JaxFunctionType = self.right.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ):
            is_less_or_equal = left_fn(
                parameters, one_row, the_draws, the_random_variables
            ) <= right_fn(parameters, one_row, the_draws, the_random_variables)
            return lax.cond(
                is_less_or_equal, lambda _: 1.0, lambda _: 0.0, operand=None
            )

        return the_jax_function


class GreaterOrEqual(ComparisonOperator):
    """
    Logical greater or equal
    """

    def __init__(self, left: ExpressionOrNumeric, right: ExpressionOrNumeric):
        """Constructor

        :param left: first arithmetic expression
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression
        :type right: biogeme.expressions.Expression
        """
        super().__init__(left, right)

    def deep_flat_copy(self) -> GreaterOrEqual:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        left_copy = self.left.deep_flat_copy()
        right_copy = self.right.deep_flat_copy()
        return type(self)(left=left_copy, right=right_copy)

    def __str__(self) -> str:
        return f'({self.left} >= {self.right})'

    def __repr__(self) -> str:
        return f'GreaterOrEqual({repr(self.left)}, {repr(self.right)})'

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        r = 1 if self.left.get_value() >= self.right.get_value() else 0
        return r

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        import jax.numpy as jnp

        left_fn: JaxFunctionType = self.left.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )
        right_fn: JaxFunctionType = self.right.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ):
            is_greater_or_equal = left_fn(
                parameters, one_row, the_draws, the_random_variables
            ) >= right_fn(parameters, one_row, the_draws, the_random_variables)
            return lax.cond(
                is_greater_or_equal, lambda _: 1.0, lambda _: 0.0, operand=None
            )

        return the_jax_function


class Less(ComparisonOperator):
    """
    Logical less
    """

    def __init__(self, left: ExpressionOrNumeric, right: ExpressionOrNumeric):
        """Constructor

        :param left: first arithmetic expression
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression
        :type right: biogeme.expressions.Expression
        """
        super().__init__(left, right)

    def deep_flat_copy(self) -> Less:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        left_copy = self.left.deep_flat_copy()
        right_copy = self.right.deep_flat_copy()
        return type(self)(left=left_copy, right=right_copy)

    def __str__(self) -> str:
        return f'({self.left} < {self.right})'

    def __repr__(self) -> str:
        return f'Less({repr(self.left)}, {repr(self.right)})'

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        r = 1 if self.left.get_value() < self.right.get_value() else 0
        return r

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        import jax.numpy as jnp

        left_fn: JaxFunctionType = self.left.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )
        right_fn: JaxFunctionType = self.right.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ):
            is_less = left_fn(
                parameters, one_row, the_draws, the_random_variables
            ) < right_fn(parameters, one_row, the_draws, the_random_variables)
            return lax.cond(is_less, lambda _: 1.0, lambda _: 0.0, operand=None)

        return the_jax_function


class Greater(ComparisonOperator):
    """
    Logical greater
    """

    def __init__(self, left: ExpressionOrNumeric, right: ExpressionOrNumeric):
        """Constructor

        :param left: first arithmetic expression
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression
        :type right: biogeme.expressions.Expression
        """
        super().__init__(left, right)

    def deep_flat_copy(self) -> Greater:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        left_copy = self.left.deep_flat_copy()
        right_copy = self.right.deep_flat_copy()
        return type(self)(left=left_copy, right=right_copy)

    def __str__(self) -> str:
        return f'({self.left} > {self.right})'

    def __repr__(self) -> str:
        return f'Greater({repr(self.left)}, {repr(self.right)})'

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        r = 1 if self.left.get_value() > self.right.get_value() else 0
        return r

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        import jax.numpy as jnp

        left_fn: JaxFunctionType = self.left.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )
        right_fn: JaxFunctionType = self.right.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ):
            is_greater = left_fn(
                parameters, one_row, the_draws, the_random_variables
            ) > right_fn(parameters, one_row, the_draws, the_random_variables)
            return lax.cond(is_greater, lambda _: 1.0, lambda _: 0.0, operand=None)

        return the_jax_function



import unittest

import biogeme.biogeme as bio
from biogeme import models
from biogeme.data.swissmetro import (
    CAR_AV_SP,
    CAR_CO_SCALED,
    CAR_TT_SCALED,
    CHOICE,
    GA,
    PURPOSE,
    SM_AV,
    SM_CO,
    SM_COST_SCALED,
    SM_TT_SCALED,
    TRAIN_AV_SP,
    TRAIN_CO,
    TRAIN_COST_SCALED,
    TRAIN_TT_SCALED,
    read_data,
)
from biogeme.expressions import Beta
from biogeme.nests import NestsForNestedLogit, OneNestForNestedLogit

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)


ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

MU = Beta('MU', 2.05, 1, None, 0)

# Additional parameter designed to estimate the bias due to choice
# based sampling
SB_TRAIN = Beta('SB_TRAIN', 0, None, None, 0)

correction = {1: SB_TRAIN, 2: 0, 3: 0}


SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)


V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}


av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of nests:
# 1: nests parameter
# 2: list of alternatives
existing = OneNestForNestedLogit(
    nest_param=MU, list_of_alternatives=[1, 3], name='existing'
)
nests = NestsForNestedLogit(tuple_of_nests=(existing,), choice_set=[1, 2, 3])

# The choice model is a nested logit, with corrections for endogenous sampling
Gi = models.get_mev_for_nested(V, av, nests)
logprob = models.logmev_endogenous_sampling(V, Gi, av, correction, CHOICE)


class test_14(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(
            database,
            logprob,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
        )
        results = biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, -5169.641515468776, 2)


if __name__ == '__main__':
    unittest.main()



"""

Factor analysis
===============

Preliminary analysis of the indicators using factor analysis

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 16:39:54 2023
"""

import sys
import pandas as pd
import numpy as np

# %%
# The following package can be installed using
#
# `pip install factor_analyzer`
#
# See https://github.com/EducationalTestingService/factor_analyzer
try:
    from factor_analyzer import FactorAnalyzer
except ModuleNotFoundError:
    print('Use "pip install factor_analyzer" to install a requested package')
    sys.exit()

# %%
# We first extract the columns containing the indicators
variables = [
    'Envir01',
    'Envir02',
    'Envir03',
    'Envir04',
    'Envir05',
    'Envir06',
    'Mobil01',
    'Mobil02',
    'Mobil03',
    'Mobil04',
    'Mobil05',
    'Mobil06',
    'Mobil07',
    'Mobil08',
    'Mobil09',
    'Mobil10',
    'Mobil11',
    'Mobil12',
    'Mobil13',
    'Mobil14',
    'Mobil15',
    'Mobil16',
    'Mobil17',
    'Mobil18',
    'Mobil19',
    'Mobil20',
    'Mobil21',
    'Mobil22',
    'Mobil23',
    'Mobil24',
    'Mobil25',
    'Mobil26',
    'Mobil27',
    'ResidCh01',
    'ResidCh02',
    'ResidCh03',
    'ResidCh04',
    'ResidCh05',
    'ResidCh06',
    'ResidCh07',
    'LifSty01',
    'LifSty02',
    'LifSty03',
    'LifSty04',
    'LifSty05',
    'LifSty06',
    'LifSty07',
    'LifSty08',
    'LifSty09',
    'LifSty10',
    'LifSty11',
    'LifSty12',
    'LifSty13',
    'LifSty14',
]

indicators = pd.read_csv('optima.dat', sep='\t', usecols=variables)

# %%
# Negative values are missing values.
indicators[indicators <= 0] = np.nan
indicators = indicators.dropna(axis=0, how='any')

# %%
# We perform the factor analysis
fa = FactorAnalyzer(rotation='varimax')
fa.fit(indicators)

# %%
# We obtain the factor loadings and label them
labeled_results = pd.DataFrame(fa.loadings_)
labeled_results.index = variables
labeled_results

# %%
# When we print the results, we keep only loadings that are 0.4 or
# higher, in absolute value.

labeled_results = labeled_results.astype('object')
labeled_results[(labeled_results <= 0.4) & (labeled_results >= -0.4)] = ''

# %%
# Results.
labeled_results



"""Unit tests for the Jax implementation of the expressions.

Michel Bierlaire
Sat Mar 29 10:56:13 2025

"""

import logging
import unittest

import jax.numpy as jnp
import numpy as np
import pandas as pd
from scipy.integrate import quad
from scipy.stats import norm

from biogeme.calculator import calculate_single_formula
from biogeme.database import Database
from biogeme.draws import DrawsManagement
from biogeme.expressions import (
    Beta,
    BinaryMax,
    BinaryMin,
    Draws,
    IntegrateNormal,
    LogLogit,
    MonteCarlo,
    NormalCdf,
    RandomVariable,
    Variable,
    cos,
    exp,
    log,
    logzero,
    sin,
)
from biogeme.expressions.power_constant import PowerConstant
from biogeme.model_elements import ModelElements
from biogeme.second_derivatives import SecondDerivativesMode

logging.basicConfig(level=logging.WARNING)


class TestBetaConstructJaxFunction(unittest.TestCase):
    def setUp(self):
        """Set up a Beta instance for testing"""
        self.beta_1 = Beta(
            name="beta_1", value=1.0, lowerbound=None, upperbound=None, status=0
        )
        self.beta_2 = Beta(
            name="beta_2", value=2.0, lowerbound=None, upperbound=None, status=0
        )

        data = pd.DataFrame(
            {"income": [1, 2, 3], "age": [10, 20, 30], "choice": [0, 1, 0]}
        )
        self.database = Database('test', dataframe=data)
        self.draws_manager = DrawsManagement(sample_size=3, number_of_draws=4)

    def test_beta(self):
        """Test if the generated JAX function correctly retrieves the parameter value."""

        parameters = {'beta_1': 0.5}  # Example parameters

        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=self.beta_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum(
            [
                0.5,
                0.5,
                0.5,
            ]
        )
        self.assertEqual(jax_1.function, expected_result)
        expected_gradient = [3]
        self.assertEqual(jax_1.gradient, expected_gradient)
        expected_hessian = [[0]]
        self.assertEqual(jax_1.hessian, expected_hessian)
        expected_bhhh = [[3]]
        self.assertEqual(jax_1.bhhh, expected_bhhh)

    def test_fixed_beta(self):

        beta_value = 0.5
        parameters = {'beta_1': beta_value}  # Example parameters
        value = 3
        fixed_beta = Beta(
            name="fixed_beta", value=value, lowerbound=None, upperbound=None, status=1
        )

        the_expression = fixed_beta + self.beta_1

        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=the_expression, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum(
            [
                beta_value + value,
                beta_value + value,
                beta_value + value,
            ]
        )
        self.assertEqual(jax_1.function, expected_result)
        expected_gradient = [3]
        self.assertEqual(jax_1.gradient, expected_gradient)
        expected_hessian = [[0]]
        self.assertEqual(jax_1.hessian, expected_hessian)
        expected_bhhh = [[3]]
        self.assertEqual(jax_1.bhhh, expected_bhhh)
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=fixed_beta, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=False,
            hessian=False,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        expected_result = sum(
            [
                value,
                value,
                value,
            ]
        )

        self.assertEqual(jax_2.function, expected_result)

    def test_variable(self):
        """Test if the generated JAX function correctly retrieves the parameter value."""

        parameters = {}
        expression_income = Variable('income')

        model_elements_income = ModelElements.from_expression_and_weight(
            log_like=expression_income, weight=None, database=self.database
        )
        jax_income = calculate_single_formula(
            model_elements=model_elements_income,
            the_betas=parameters,
            gradient=False,
            hessian=False,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1, 2, 3])
        self.assertEqual(jax_income.function, expected_result)
        expected_gradient = None
        self.assertEqual(jax_income.gradient, expected_gradient)
        expected_hessian = None
        self.assertEqual(jax_income.hessian, expected_hessian)

        expression_age = Variable('age')

        model_elements_age = ModelElements.from_expression_and_weight(
            log_like=expression_age, weight=None, database=self.database
        )

        jax_age = calculate_single_formula(
            model_elements=model_elements_age,
            the_betas=parameters,
            gradient=False,
            hessian=False,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([10, 20, 30])
        self.assertEqual(jax_age.function, expected_result)

        expression_choice = Variable('choice')

        model_elements_choice = ModelElements.from_expression_and_weight(
            log_like=expression_choice, weight=None, database=self.database
        )

        jax_choice = calculate_single_formula(
            model_elements=model_elements_choice,
            the_betas=parameters,
            gradient=False,
            hessian=False,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([0, 1, 0])
        self.assertEqual(jax_choice.function, expected_result)

    def test_plus(self):
        parameters = {'beta_1': 0.9, 'beta_2': 1.5}  # Example parameters
        expression_1 = self.beta_1 + self.beta_2
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([0.9 + 1.5, 0.9 + 1.5, 0.9 + 1.5])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient = jnp.array([3, 3])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient))
        expected_hessian = jnp.array([[0, 0], [0, 0]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian))
        expected_bhhh = jnp.array([[3.0, 3.0], [3.0, 3.0]])
        self.assertEqual(jax_1.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_1.bhhh, expected_bhhh))

        parameters = {'beta_2': 1.5}
        expression_2 = self.beta_2 + Variable('age')
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1.5 + 10, 1.5 + 20, 1.5 + 30])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))

        expected_gradient = jnp.array([3])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient))
        expected_hessian = jnp.array([[0]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian))
        expected_bhhh = jnp.array([[3]])
        self.assertEqual(jax_2.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_2.bhhh, expected_bhhh))

        expression_3 = Variable('income') + self.beta_2 + Variable('age')
        model_elements_3 = ModelElements.from_expression_and_weight(
            log_like=expression_3, weight=None, database=self.database
        )
        jax_3 = calculate_single_formula(
            model_elements=model_elements_3,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1 + 1.5 + 10, 2 + 1.5 + 20, 3 + 1.5 + 30])
        self.assertTrue(jnp.allclose(jax_3.function, expected_result))
        expected_gradient = jnp.array([3])
        self.assertTrue(jnp.allclose(jax_3.gradient, expected_gradient))
        expected_hessian = jnp.array([[0]])
        self.assertEqual(jax_3.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_3.hessian, expected_hessian))
        expected_bhhh = jnp.array([[3]])
        self.assertEqual(jax_3.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_3.bhhh, expected_bhhh))

    def test_minus(self):
        parameters = {'beta_1': 0.9, 'beta_2': 1.5}  # Example parameters
        expression_1 = self.beta_1 - self.beta_2
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([0.9 - 1.5, 0.9 - 1.5, 0.9 - 1.5])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient = jnp.array([3, -3])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient))
        expected_bhhh = np.array([[3.0, -3.0], [-3.0, 3.0]])
        self.assertTrue(jnp.allclose(jax_1.bhhh, expected_bhhh))
        expected_hessian = jnp.array([[0, 0], [0, 0]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian))

        parameters = {'beta_2': 1.5}
        expression_2 = self.beta_2 - Variable('age')
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1.5 - 10, 1.5 - 20, 1.5 - 30])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))
        expected_gradient = jnp.array([3])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient))
        expected_bhhh = jnp.array([[3]])
        self.assertTrue(jnp.allclose(jax_2.bhhh, expected_bhhh))
        expected_hessian = jnp.array([[0]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian))

        expression_3 = Variable('income') - self.beta_2 - Variable('age')
        model_elements_3 = ModelElements.from_expression_and_weight(
            log_like=expression_3, weight=None, database=self.database
        )
        jax_3 = calculate_single_formula(
            model_elements=model_elements_3,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1 - 1.5 - 10, 2 - 1.5 - 20, 3 - 1.5 - 30])
        self.assertTrue(jnp.allclose(jax_3.function, expected_result))
        expected_gradient = jnp.array([-3])
        self.assertTrue(jnp.allclose(jax_3.gradient, expected_gradient))
        expected_bhhh = jnp.array([[3]])
        self.assertTrue(jnp.allclose(jax_3.bhhh, expected_bhhh))
        expected_hessian = jnp.array([[0]])
        self.assertEqual(jax_3.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_3.hessian, expected_hessian))

    def test_times(self):
        parameters = {'beta_1': 0.9, 'beta_2': 1.5}  # Example parameters
        expression_1 = self.beta_1 * self.beta_2
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([0.9 * 1.5, 0.9 * 1.5, 0.9 * 1.5])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient = jnp.array([1.5 * 3, 0.9 * 3])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient))
        expected_hessian = jnp.array([[0, 3], [3, 0]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian))
        expected_bhhh = jnp.array([[6.75, 4.05], [4.05, 2.43]])
        self.assertEqual(jax_1.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_1.bhhh, expected_bhhh))

        parameters = {'beta_2': 1.5}
        expression_2 = self.beta_2 * Variable('age')
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )

        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        expected_result = sum([1.5 * 10, 1.5 * 20, 1.5 * 30])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))
        expected_gradient = jnp.array([10 + 20 + 30])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient))
        expected_hessian = jnp.array([[0]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian))
        expected_bhhh = jnp.array([[1400]])
        self.assertEqual(jax_2.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_2.bhhh, expected_bhhh))

        expression_3 = Variable('income') * self.beta_2 * Variable('age')
        model_elements_3 = ModelElements.from_expression_and_weight(
            log_like=expression_3, weight=None, database=self.database
        )

        jax_3 = calculate_single_formula(
            model_elements=model_elements_3,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1 * 1.5 * 10, 2 * 1.5 * 20, 3 * 1.5 * 30])
        self.assertTrue(jnp.allclose(jax_3.function, expected_result))
        expected_gradient = jnp.array([1 * 10 + 2 * 20 + 3 * 30])
        self.assertTrue(jnp.allclose(jax_3.gradient, expected_gradient))
        expected_hessian = jnp.array([[0]])
        self.assertEqual(jax_3.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_3.hessian, expected_hessian))
        expected_bhhh = jnp.array([[9800]])
        self.assertEqual(jax_3.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_3.bhhh, expected_bhhh))

    def test_divide(self):
        parameters = {'beta_1': 0.9, 'beta_2': 1.5}  # Example parameters
        expression_1 = self.beta_1 / self.beta_2
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([0.9 / 1.5, 0.9 / 1.5, 0.9 / 1.5])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient = jnp.array([3 / 1.5, -3 * 0.9 / (1.5 * 1.5)])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient))
        expected_hessian = jnp.array(
            [[0, -3 / 1.5**2], [-3 / 1.5**2, 6 * 0.9 / 1.5**3]]
        )
        self.assertEqual(jax_1.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian))
        expected_bhhh = jnp.array([[4 / 3, -0.8], [-0.8, 0.48]])
        self.assertEqual(jax_1.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_1.bhhh, expected_bhhh))

        parameters = {'beta_2': 1.5}
        expression_2 = self.beta_2 / Variable('age')
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1.5 / 10, 1.5 / 20, 1.5 / 30])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))
        expected_gradient = jnp.array([1.0 / 10 + 1.0 / 20 + 1.0 / 30])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient))
        expected_hessian = jnp.array([[0]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian))
        expected_bhhh = jnp.array([[0.01361111]])
        self.assertEqual(jax_2.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_2.bhhh, expected_bhhh))

        expression_3 = Variable('income') / self.beta_2 / Variable('age')
        model_elements_3 = ModelElements.from_expression_and_weight(
            log_like=expression_3, weight=None, database=self.database
        )
        jax_3 = calculate_single_formula(
            model_elements=model_elements_3,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1 / 1.5 / 10, 2 / 1.5 / 20, 3 / 1.5 / 30])
        self.assertTrue(jnp.allclose(jax_3.function, expected_result))
        expected_gradient = jnp.array(
            [-1.0 / (10 * 1.5**2) - 2.0 / (20 * 1.5**2) - 3.0 / (30 * 1.5**2)]
        )
        self.assertTrue(jnp.allclose(jax_3.gradient, expected_gradient))
        expected_hessian = jnp.array(
            [[2 * 1 / (10 * 1.5**3) + 2 * 2 / (20 * 1.5**3) + 2 * 3 / (30 * 1.5**3)]]
        )
        self.assertEqual(jax_3.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_3.hessian, expected_hessian))
        expected_bhhh = jnp.array([[0.00592593]])
        self.assertEqual(jax_3.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_3.bhhh, expected_bhhh))

    def test_power(self):
        parameters = {'beta_1': 0.9, 'beta_2': 1.5}  # Example parameters
        expression_1 = self.beta_1**self.beta_2
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        expected_result = sum([0.9**1.5, 0.9**1.5, 0.9**1.5])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient = jnp.array(
            [
                3
                * parameters['beta_2']
                * parameters['beta_1'] ** (parameters['beta_2'] - 1),  # ∂f/∂β₁
                3
                * np.log(parameters['beta_1'])
                * parameters['beta_1'] ** parameters['beta_2'],  # ∂f/∂β₂
            ]
        )
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient))
        expected_hessian = jnp.array(
            [
                [
                    3
                    * parameters['beta_2']
                    * (parameters['beta_2'] - 1)
                    * parameters['beta_1'] ** (parameters['beta_2'] - 2),
                    3
                    * parameters['beta_1'] ** (parameters['beta_2'] - 1)
                    * (1 + parameters['beta_2'] * np.log(parameters['beta_1'])),
                ],
                [
                    3
                    * parameters['beta_1'] ** (parameters['beta_2'] - 1)
                    * (1 + parameters['beta_2'] * np.log(parameters['beta_1'])),
                    3
                    * (np.log(parameters['beta_1'])) ** 2
                    * parameters['beta_1'] ** parameters['beta_2'],
                ],
            ]
        )
        self.assertEqual(jax_1.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian))
        expected_bhhh = jnp.array([[6.075, -0.38403916], [-0.38403916, 0.02427755]])
        self.assertEqual(jax_1.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_1.bhhh, expected_bhhh))

        parameters = {'beta_2': 1.5}
        expression_2 = self.beta_2 ** Variable('age')
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        expected_result = sum([1.5**10, 1.5**20, 1.5**30])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))
        expected_gradient = jnp.array(
            [
                10 * parameters['beta_2'] ** 9
                + 20 * parameters['beta_2'] ** 19
                + 30 * parameters['beta_2'] ** 29
            ]
        )
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient))

        expected_hessian = jnp.array(
            [
                [
                    10 * 9 * parameters['beta_2'] ** 8
                    + 20 * 19 * parameters['beta_2'] ** 18
                    + 30 * 29 * parameters['beta_2'] ** 28
                ]
            ]
        )
        self.assertEqual(jax_2.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian))
        expected_bhhh = jnp.array([[1.4709354e13]])
        self.assertEqual(jax_2.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_2.bhhh, expected_bhhh))

    def test_min(self):
        parameters = {'beta_1': 0.9, 'beta_2': 1.5}  # Example parameters
        expression_1 = BinaryMin(self.beta_1, self.beta_2)
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([0.9, 0.9, 0.9])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))

        expected_gradient = jnp.array([3, 0])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient))

        expected_hessian = jnp.array([[0, 0], [0, 0]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian))

        expected_bhhh = jnp.array([[3, 0], [0, 0]])
        self.assertEqual(jax_1.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_1.bhhh, expected_bhhh))

        parameters = {'beta_2': 1.5}
        expression_2 = BinaryMin(self.beta_2, Variable('age'))
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1.5, 1.5, 1.5])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))
        expected_gradient = jnp.array([3])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient))

        expected_hessian = jnp.array([[0]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian))

        expected_bhhh = jnp.array([[3]])
        self.assertEqual(jax_2.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_2.bhhh, expected_bhhh))

    def test_max(self):
        parameters = {'beta_1': 0.9, 'beta_2': 1.5}  # Example parameters
        expression_1 = BinaryMax(self.beta_1, self.beta_2)
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        expected_result = sum([1.5, 1.5, 1.5])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient = jnp.array([0, 3])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient))

        expected_hessian = jnp.array([[0, 0], [0, 0]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian))

        expected_bhhh = jnp.array([[0, 0], [0, 3]])
        self.assertEqual(jax_1.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_1.bhhh, expected_bhhh))

        parameters = {'beta_2': 1.5}
        expression_2 = BinaryMax(self.beta_2, Variable('age'))
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        expected_result = sum([10, 20, 30])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))

        expected_gradient = jnp.array([0])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient))

        expected_hessian = jnp.array([[0]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian))

        expected_bhhh = jnp.array([[0]])
        self.assertEqual(jax_2.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_2.bhhh, expected_bhhh))

    def test_and(self):
        parameters = {'beta_1': 0, 'beta_2': 1.5}  # Example parameters
        expression_1 = self.beta_1 & self.beta_2
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([0, 0, 0])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient = jnp.array([0, 0])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient))

        expected_hessian = jnp.array([[0, 0], [0, 0]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian))

        expected_bhhh = jnp.array([[0, 0], [0, 0]])
        self.assertEqual(jax_1.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_1.bhhh, expected_bhhh))

        parameters = {'beta_1': 1.5}  # Example parameters
        expression_2 = self.beta_1 * (Variable('age') & Variable('income'))
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1.5, 1.5, 1.5])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))
        expected_gradient = jnp.array([3])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient))

        expected_hessian = jnp.array([[0]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian))
        expected_bhhh = jnp.array([[3]])
        self.assertEqual(jax_2.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_2.bhhh, expected_bhhh))

        parameters = {'beta_1': 1, 'beta_2': 1.5}  # Example parameters
        expression_3 = self.beta_1 & self.beta_2
        model_elements_3 = ModelElements.from_expression_and_weight(
            log_like=expression_3, weight=None, database=self.database
        )
        jax_3 = calculate_single_formula(
            model_elements=model_elements_3,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1, 1, 1])
        self.assertTrue(jnp.allclose(jax_3.function, expected_result))
        expected_gradient = jnp.array([0, 0])
        self.assertTrue(jnp.allclose(jax_3.gradient, expected_gradient))

        expected_hessian = jnp.array([[0, 0], [0, 0]])
        self.assertEqual(jax_3.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_3.hessian, expected_hessian))

        expected_bhhh = jnp.array([[0, 0], [0, 0]])
        self.assertEqual(jax_3.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_3.bhhh, expected_bhhh))

    def test_or(self):
        parameters = {'beta_1': 0, 'beta_2': 1.5}  # Example parameters
        expression_1 = self.beta_1 | self.beta_2
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1, 1, 1])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient = jnp.array([0, 0])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient))

        expected_hessian = jnp.array([[0, 0], [0, 0]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian))

        expected_bhhh = jnp.array([[0, 0], [0, 0]])
        self.assertEqual(jax_1.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_1.bhhh, expected_bhhh))

        parameters = {'beta_1': 1.5}  # Example parameters
        expression_2 = self.beta_1 * (Variable('age') | Variable('income'))
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1.5, 1.5, 1.5])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))
        expected_gradient = jnp.array([3])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient))

        expected_hessian = jnp.array([[0]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian))

        expected_bhhh = jnp.array([[3]])
        self.assertEqual(jax_2.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_2.bhhh, expected_bhhh))

    def test_unary_minus(self):
        parameters = {'beta_1': -1}  # Example parameters
        expression_1 = -self.beta_1
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1, 1, 1])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient = jnp.array([-3])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient))

        expected_hessian = jnp.array([[0]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian))

        expected_bhhh = jnp.array([[3]])
        self.assertEqual(jax_1.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax_1.bhhh, expected_bhhh))

        expression_2 = -Variable('age')
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )

        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas={},
            gradient=False,
            hessian=False,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([-10, -20, -30])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))

    def test_normal_cdf(self):
        value_1 = 0
        cdf_1 = norm.cdf(value_1)
        pdf_1 = norm.pdf(value_1)
        value_2 = -1
        cdf_2 = norm.cdf(value_2)
        pdf_2 = norm.pdf(value_2)
        parameters = {'beta_1': value_1}
        expression_1 = NormalCdf(self.beta_1)
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([cdf_1, cdf_1, cdf_1])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient_1 = jnp.array([3 * pdf_1])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient_1))
        expected_hessian_1 = jnp.array([[-3 * value_1 * pdf_1]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian_1.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian_1))

        expected_bhhh_1 = jnp.array([[0.47746485]])
        self.assertEqual(jax_1.bhhh.shape, expected_bhhh_1.shape)
        self.assertTrue(jnp.allclose(jax_1.bhhh, expected_bhhh_1))

        parameters = {'beta_2': value_2}
        expression_2 = NormalCdf(self.beta_2)
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([cdf_2, cdf_2, cdf_2])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))
        expected_gradient_2 = jnp.array([3 * pdf_2])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient_2))
        expected_hessian_2 = jnp.array([[-3 * value_2 * pdf_2]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian_2.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian_2))

        expected_bhhh_2 = jnp.array([[0.17564951]])
        self.assertEqual(jax_2.bhhh.shape, expected_bhhh_2.shape)
        self.assertTrue(jnp.allclose(jax_2.bhhh, expected_bhhh_2))

    def test_exp(self):
        value_1 = 0
        exp_1 = np.exp(value_1)
        value_2 = -1
        exp_2 = np.exp(value_2)
        parameters = {'beta_1': value_1}  # Example parameters
        expression_1 = exp(self.beta_1)
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([exp_1, exp_1, exp_1])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient_1 = jnp.array([3 * exp_1])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient_1))
        expected_hessian_1 = jnp.array([[3]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian_1.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian_1))
        expected_bhhh_1 = jnp.array([[3]])
        self.assertEqual(jax_1.bhhh.shape, expected_bhhh_1.shape)
        self.assertTrue(jnp.allclose(jax_1.bhhh, expected_bhhh_1))

        parameters = {'beta_2': value_2}  # Example parameters

        expression_2 = exp(self.beta_2)
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([exp_2, exp_2, exp_2])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))

    def test_sin(self):
        value_1 = 0
        sin_1 = np.sin(value_1)
        value_2 = -1
        sin_2 = np.sin(value_2)
        parameters = {'beta_1': value_1}  # Example parameters

        # Test sin(beta_1)
        expression_1 = sin(self.beta_1)
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([sin_1, sin_1, sin_1])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))

        expected_gradient_1 = jnp.array([3 * np.cos(value_1)])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient_1))

        expected_hessian_1 = jnp.array([[-3 * np.sin(value_1)]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian_1.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian_1))

        # Test sin(beta_2)
        parameters = {'beta_2': value_2}
        expression_2 = sin(self.beta_2)
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )

        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([sin_2, sin_2, sin_2])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))

        expected_gradient_2 = jnp.array([3 * np.cos(value_2)])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient_2))

        expected_hessian_2 = jnp.array([[-3 * np.sin(value_2)]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian_2.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian_2))

    def test_cos(self):
        value_1 = 0
        cos_1 = np.cos(value_1)
        value_2 = -1
        cos_2 = np.cos(value_2)
        parameters = {'beta_1': value_1}  # Example parameters
        expression_1 = cos(self.beta_1)
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([cos_1, cos_1, cos_1])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        parameters = {'beta_2': value_2}
        expression_2 = cos(self.beta_2)
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )

        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([cos_2, cos_2, cos_2])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))

    def test_log(self):
        value_1 = 1
        log_1 = np.log(value_1)
        value_2 = 2
        log_2 = np.log(value_2)
        parameters = {'beta_1': value_1}  # Example parameters
        expression_1 = log(self.beta_1)
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([log_1, log_1, log_1])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient_1 = jnp.array([3 * (1 / value_1)])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient_1))
        expected_hessian_1 = jnp.array([[-3 * (1 / value_1**2)]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian_1.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian_1))

        parameters = {'beta_2': value_2}
        expression_2 = log(self.beta_2)
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([log_2, log_2, log_2])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))
        expected_gradient_2 = jnp.array([3 * (1 / value_2)])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient_2))
        expected_hessian_2 = jnp.array([[-3 * (1 / value_2**2)]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian_2.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian_2))

    def test_log_zero(self):
        value_1 = 0
        log_1 = 0
        value_2 = 2
        log_2 = np.log(value_2)
        parameters = {'beta_1': value_1}  # Example parameters
        expression_1 = logzero(self.beta_1)
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([log_1, log_1, log_1])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        expected_gradient_1 = jnp.array([0.0])
        self.assertTrue(jnp.allclose(jax_1.gradient, expected_gradient_1))
        expected_hessian_1 = jnp.array([[0.0]])
        self.assertEqual(jax_1.hessian.shape, expected_hessian_1.shape)
        self.assertTrue(jnp.allclose(jax_1.hessian, expected_hessian_1))

        parameters = {'beta_2': value_2}
        expression_2 = logzero(self.beta_2)
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([log_2, log_2, log_2])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))
        expected_gradient_2 = jnp.array([3 * (1 / value_2)])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient_2))
        expected_hessian_2 = jnp.array([[-3 * (1 / value_2**2)]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian_2.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian_2))

    def test_power_constant(self):
        parameters = {'beta_1': -2}  # Example parameters

        expression_1 = PowerConstant(self.beta_1, 2)
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1, weight=None, database=self.database
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([4, 4, 4])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))

        expression_2 = PowerConstant(self.beta_1, 3)
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2, weight=None, database=self.database
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([-8, -8, -8])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))
        parameters = {'beta_2': 0}  # Example parameters

        expression_3 = PowerConstant(self.beta_2, -3.1)
        model_elements_3 = ModelElements.from_expression_and_weight(
            log_like=expression_3, weight=None, database=self.database
        )
        jax_3 = calculate_single_formula(
            model_elements=model_elements_3,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=True,
        )
        expected_result = sum([0, 0, 0])
        self.assertTrue(jnp.allclose(jax_3.function, expected_result))

    def test_integrate(self):
        beta = Beta('beta', 1, -10, 10, 0)
        omega = RandomVariable('omega')
        integral = IntegrateNormal(omega - omega + beta, name='omega')
        model_elements = ModelElements.from_expression_and_weight(
            log_like=integral,
            weight=None,
            database=self.database,
        )
        jax_1 = calculate_single_formula(
            model_elements=model_elements,
            the_betas={'beta': 1.0},
            gradient=False,
            hessian=False,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1, 1, 1])
        self.assertTrue(jnp.allclose(jax_1.function, expected_result))
        jax_2 = calculate_single_formula(
            model_elements=model_elements,
            the_betas={'beta': 1.0},
            gradient=True,
            hessian=False,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        expected_result = sum([1, 1, 1])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))
        expected_gradient_2 = jnp.array([3])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient_2))
        jax_3 = calculate_single_formula(
            model_elements=model_elements,
            the_betas={'beta': 1.0},
            gradient=True,
            hessian=True,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        expected_result = sum([1, 1, 1])
        self.assertTrue(jnp.allclose(jax_3.function, expected_result))
        expected_gradient_3 = jnp.array([3])
        self.assertTrue(jnp.allclose(jax_3.gradient, expected_gradient_3))
        expected_hessian_3 = jnp.array([0])
        self.assertTrue(jnp.allclose(jax_3.hessian, expected_hessian_3))

        jax_4 = calculate_single_formula(
            model_elements=model_elements,
            the_betas={'beta': 1.0},
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([1, 1, 1])
        self.assertTrue(jnp.allclose(jax_4.function, expected_result))
        expected_gradient_4 = jnp.array([3])
        self.assertTrue(jnp.allclose(jax_4.gradient, expected_gradient_4))
        expected_hessian_4 = jnp.array([0])
        self.assertTrue(jnp.allclose(jax_4.hessian, expected_hessian_4))
        expected_bhhh_4 = jnp.array([3])
        self.assertTrue(jnp.allclose(jax_4.bhhh, expected_bhhh_4))

    def test_integrate_2(self):

        value_beta_1 = 1
        value_beta_2 = 1
        beta_1 = Beta('beta_1', value_beta_1, -10, 10, 0)
        beta_2 = Beta('beta_2', value_beta_2, -10, 10, 0)
        beta_parameters = {'beta_1': value_beta_1, 'beta_2': value_beta_2}

        # Integral using Monte Carlo
        draws = Draws('draws', 'NORMAL_HALTON3')
        formula_mc = beta_1 * beta_2 * exp(draws)
        integral_mc = MonteCarlo(formula_mc)
        model_elements_mc = ModelElements.from_expression_and_weight(
            log_like=integral_mc,
            weight=None,
            database=self.database,
            number_of_draws=1000000,
        )
        jax_mc = calculate_single_formula(
            model_elements=model_elements_mc,
            the_betas=beta_parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        # Integral using scipy
        def f(x):
            x = np.clip(x, a_min=None, a_max=100)  # Prevent exp(>100)
            return value_beta_1 * value_beta_2 * np.exp(x)

        def integrand(x):
            return f(x) * norm.pdf(x)

        result, error = quad(integrand, -np.inf, np.inf)

        omega = RandomVariable('omega')
        formula_rv = beta_1 * beta_2 * exp(omega)
        integral_rv = IntegrateNormal(formula_rv, name='omega')
        model_elements = ModelElements.from_expression_and_weight(
            log_like=integral_rv,
            weight=None,
            database=self.database,
        )

        jax_rv = calculate_single_formula(
            model_elements=model_elements,
            the_betas=beta_parameters,
            gradient=False,
            hessian=False,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        expected_result = sum([result, result, result])
        self.assertTrue(jnp.allclose(jax_rv.function, expected_result))

        jax_2 = calculate_single_formula(
            model_elements=model_elements,
            the_betas={'beta': 1.0},
            gradient=True,
            hessian=False,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        expected_result = sum([result, result, result])
        self.assertTrue(jnp.allclose(jax_2.function, expected_result))

        self.assertTrue(jnp.allclose(jax_2.gradient, jax_mc.gradient, rtol=1e-3))

        jax_3 = calculate_single_formula(
            model_elements=model_elements,
            the_betas={'beta': 1.0},
            gradient=True,
            hessian=True,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([result, result, result])
        self.assertTrue(jnp.allclose(jax_3.function, expected_result))

        self.assertTrue(jnp.allclose(jax_3.gradient, jax_mc.gradient, rtol=1e-3))
        self.assertTrue(jnp.allclose(jax_3.hessian, jax_mc.hessian, rtol=1e-3))

        jax_4 = calculate_single_formula(
            model_elements=model_elements,
            the_betas={'beta': 1.0},
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_result = sum([result, result, result])
        self.assertTrue(jnp.allclose(jax_4.function, expected_result))

        self.assertTrue(jnp.allclose(jax_4.gradient, jax_mc.gradient, rtol=1e-3))
        self.assertTrue(jnp.allclose(jax_4.hessian, jax_mc.hessian, rtol=1e-3))
        self.assertTrue(jnp.allclose(jax_4.bhhh, jax_mc.bhhh, rtol=1e-3))

    def test_monte_carlo(self):
        random_term = Draws('random_term', 'UNIFORM')
        second_term = Draws('second_term', 'NORMAL')
        expression_1 = MonteCarlo(random_term + second_term)
        model_elements_1 = ModelElements.from_expression_and_weight(
            log_like=expression_1,
            weight=None,
            database=self.database,
            number_of_draws=10,
        )

        expected_results = []

        jax_1 = calculate_single_formula(
            model_elements=model_elements_1,
            the_betas={},
            gradient=False,
            hessian=False,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        for obs in model_elements_1.draws_management.draws:
            integral = 0
            for draw in obs:
                integral += draw[0] + draw[1]
            expected_results.append(integral / len(obs))
        self.assertTrue(jnp.allclose(jax_1.function, sum(expected_results)))

        parameters = {'beta_1': 0.9}  # Example parameters
        expression_2 = MonteCarlo(self.beta_1 * random_term)
        expected_results = []
        expected_gradient = []
        expected_bhhh = []
        model_elements_2 = ModelElements.from_expression_and_weight(
            log_like=expression_2,
            weight=None,
            database=self.database,
            number_of_draws=10,
        )
        jax_2 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        for obs in model_elements_2.draws_management.draws:
            integral = 0
            for draw in obs:
                integral += draw[0]
            expected_results.append(0.9 * integral / len(obs))
            expected_gradient.append(integral / len(obs))
            expected_bhhh.append(integral * integral / (len(obs) * len(obs)))
        self.assertTrue(jnp.allclose(jax_2.function, sum(expected_results)))
        expected_gradient_2 = jnp.array([sum(expected_gradient)])
        self.assertTrue(jnp.allclose(jax_2.gradient, expected_gradient_2))
        expected_hessian_2 = jnp.array([[0]])
        self.assertEqual(jax_2.hessian.shape, expected_hessian_2.shape)
        self.assertTrue(jnp.allclose(jax_2.hessian, expected_hessian_2))
        expected_bhhh_2 = jnp.array([[sum(expected_bhhh)]])
        self.assertEqual(jax_2.bhhh.shape, expected_bhhh_2.shape)
        self.assertTrue(jnp.allclose(jax_2.bhhh, expected_bhhh_2))

        # Test: gradient=True, hessian=False, bhhh=False
        jax_3 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=False,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        self.assertTrue(jnp.allclose(jax_3.function, sum(expected_results)))
        self.assertTrue(jnp.allclose(jax_3.gradient, expected_gradient_2))
        self.assertIsNone(jax_3.hessian)
        self.assertIsNone(jax_3.bhhh)

        # Test: gradient=True, hessian=True, bhhh=False
        jax_4 = calculate_single_formula(
            model_elements=model_elements_2,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        self.assertTrue(jnp.allclose(jax_4.function, sum(expected_results)))
        self.assertTrue(jnp.allclose(jax_4.gradient, expected_gradient_2))
        self.assertTrue(jnp.allclose(jax_4.hessian, expected_hessian_2))
        self.assertIsNone(jax_4.bhhh)

    def test_logit_1(self):
        value_1 = 0
        value_2 = 0
        parameters = {'beta_1': value_1, 'beta_2': value_2}
        beta_1 = Beta('beta_1', 0, None, None, 0)
        beta_2 = Beta('beta_2', 0, None, None, 0)
        utilities = {12: beta_1, 23: beta_2}
        the_logit = LogLogit(utilities, None, 12)
        expected_f = 3 * np.log(0.5)
        expected_g = np.array([1.5, -1.5])
        expected_h = np.array([[-0.75, 0.75], [0.75, -0.75]])
        expected_bhhh = 3 * np.outer([0.5, -0.5], [0.5, -0.5])
        model_elements = ModelElements.from_expression_and_weight(
            log_like=the_logit, weight=None, database=self.database
        )
        jax = calculate_single_formula(
            model_elements=model_elements,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        self.assertTrue(jnp.allclose(jax.function, expected_f))
        self.assertTrue(jnp.allclose(jax.gradient, expected_g))
        self.assertEqual(jax.hessian.shape, expected_h.shape)
        self.assertTrue(jnp.allclose(jax.hessian, expected_h))
        self.assertEqual(jax.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax.bhhh, expected_bhhh))

    def test_logit_2(self):
        value_1 = 0.1
        value_2 = 0.2
        parameters = {'beta_1': value_1, 'beta_2': value_2}
        beta_1 = Beta('beta_1', 0, None, None, 0)
        beta_2 = Beta('beta_2', 0, None, None, 0)
        utilities = {12: beta_1, 23: beta_2}
        the_logit = LogLogit(utilities, None, 12)
        expected_function = -2.2331900596618652
        expected_gradient = np.array([1.5749376, -1.5749376])
        expected_hessian = np.array([[-0.7481281, 0.7481281], [0.7481281, -0.7481281]])
        expected_bhhh = np.array([[0.82680964, -0.8268095], [-0.8268095, 0.8268094]])
        model_elements = ModelElements.from_expression_and_weight(
            log_like=the_logit, weight=None, database=self.database
        )
        jax = calculate_single_formula(
            model_elements=model_elements,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        self.assertTrue(jnp.allclose(jax.function, expected_function))
        self.assertTrue(jnp.allclose(jax.gradient, expected_gradient))
        self.assertEqual(jax.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax.hessian, expected_hessian))
        self.assertEqual(jax.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax.bhhh, expected_bhhh))

    def test_logit_3(self):
        value_1 = 0.1
        value_2 = 0.2
        parameters = {'beta_1': value_1, 'beta_2': value_2}
        beta_1 = Beta('beta_1', 0, None, None, 0)
        beta_2 = Beta('beta_2', 0, None, None, 0)
        utilities = {12: beta_1, 23: beta_2}
        av = {12: 1, 23: 0}
        the_logit = LogLogit(utilities, av, 12)
        model_elements = ModelElements.from_expression_and_weight(
            log_like=the_logit, weight=None, database=self.database
        )
        jax = calculate_single_formula(
            model_elements=model_elements,
            the_betas=parameters,
            gradient=True,
            hessian=True,
            bhhh=True,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected_function = 0.0
        expected_gradient = np.array([0.0, 0.0])
        expected_hessian = np.array([[0.0, 0.0], [0.0, 0.0]])
        expected_bhhh = np.array([[0.0, 0.0], [0.0, 0.0]])
        self.assertTrue(jnp.allclose(jax.function, expected_function))
        self.assertTrue(jnp.allclose(jax.gradient, expected_gradient))
        self.assertEqual(jax.hessian.shape, expected_hessian.shape)
        self.assertTrue(jnp.allclose(jax.hessian, expected_hessian))
        self.assertEqual(jax.bhhh.shape, expected_bhhh.shape)
        self.assertTrue(jnp.allclose(jax.bhhh, expected_bhhh))


if __name__ == '__main__':
    unittest.main()
